import math
from typing import List, Tuple

import torch
import torch.autograd as autograd
import torch.nn as nn
import torchmetrics
from torch.utils.data import DataLoader

from .basic_trainer import BasicTrainer


def gradient_penalty(model, real, fake, label=None):
    device = real.device
    real = real.data
    fake = fake.data
    alpha = torch.rand(real.size(0), *([1] * (len(real.shape) - 1))).to(device)
    inputs = alpha * real + ((1 - alpha) * fake)
    inputs.requires_grad_()

    if label is None:
        outputs = model(inputs)
    else:
        outputs = model(inputs, label)

    gradient = autograd.grad(outputs=outputs,
                             inputs=inputs,
                             grad_outputs=torch.ones_like(outputs).to(device),
                             create_graph=True,
                             retain_graph=True,
                             only_inputs=True)[0]

    gradient = gradient.flatten(1)
    return ((gradient.norm(2, dim=1) - 1)**2).mean()


class GANTrainer(BasicTrainer):
    r'''
    This class provide the implementation for WGAN-GP. It trains a zero-sum game between the generator and the discriminator, just like the traditional generative networks. The generator is optimized to generate simulation samples that are indistinguishable by the discriminator, and the discriminator is optimized to discriminate false samples generated by the generator. Compared with vanilla GAN, with WGAN-GP we can improve the stability of learning, get rid of problems like mode collapse, and provide meaningful learning curves useful for debugging and hyperparameter searches. Thus, existing work typically uses WGAN-GP to generate simulated EEG signals. For more details, please refer to the following information. 

    - Paper: Gulrajani I, Ahmed F, Arjovsky M, et al. Improved training of wasserstein gans[J]. Advances in neural information processing systems, 2017, 30.
    - URL: https://arxiv.org/abs/1704.00028
    - Related Project: https://github.com/eriklindernoren/PyTorch-GAN

    .. code-block:: python
        
        g_model = BGenerator(in_channels=128)
        d_model = BDiscriminator(in_channels=4)
        trainer = GANTrainer(generator, discriminator)
        trainer.fit(train_loader, val_loader)
        trainer.test(test_loader)

    The class provides the following hook functions for inserting additional implementations in the training, validation and testing lifecycle:

    - :obj:`before_training_epoch`: executed before each epoch of training starts
    - :obj:`before_training_step`: executed before each batch of training starts
    - :obj:`on_training_step`: the training process for each batch
    - :obj:`after_training_step`: execute after the training of each batch
    - :obj:`after_training_epoch`: executed after each epoch of training
    - :obj:`before_validation_epoch`: executed before each round of validation starts
    - :obj:`before_validation_step`: executed before the validation of each batch
    - :obj:`on_validation_step`: validation process for each batch
    - :obj:`after_validation_step`: executed after the validation of each batch
    - :obj:`after_validation_epoch`: executed after each round of validation
    - :obj:`before_test_epoch`: executed before each round of test starts
    - :obj:`before_test_step`: executed before the test of each batch
    - :obj:`on_test_step`: test process for each batch
    - :obj:`after_test_step`: executed after the test of each batch
    - :obj:`after_test_epoch`: executed after each round of test

    If you want to customize some operations, you just need to inherit the class and override the hook function:

    .. code-block:: python

        class MyGANTrainer(GANTrainer):
            def before_training_epoch(self, epoch_id: int, num_epochs: int):
                # Do something here.
                super().before_training_epoch(epoch_id, num_epochs)
    
    If you want to use multiple GPUs for parallel computing, you need to specify the GPU indices you want to use in the python file:
    
    .. code-block:: python

        trainer = GANTrainer(generator, discriminator, device_ids=[1, 2, 7])
        trainer.fit(train_loader, val_loader)
        trainer.test(test_loader)

    Then, you can use the :obj:`torch.distributed.launch` or :obj:`torchrun` to run your python file.

    .. code-block:: shell

        python -m torch.distributed.launch \
            --nproc_per_node=3 \
            --nnodes=1 \
            --node_rank=0 \
            --master_addr="localhost" \
            --master_port=2345 \
            your_python_file.py

    Here, :obj:`nproc_per_node` is the number of GPUs you specify.

    Args:
        generator (nn.Module): The generator model for EEG signal generation, whose inputs are Gaussian distributed random vectors, outputs are generated EEG signals. The dimensions of the input vector should be defined on the :obj:`in_channel` attribute. The output layer does not need to have a softmax activation function.
        discriminator (nn.Module): The discriminator model to determine whether the EEG signal is real or generated, and the dimension of its output should be equal to the one (i.e., the score to distinguish the real and the fake). The output layer does not need to have a sigmoid activation function.
        generator_lr (float): The learning rate of the generator. (defualt: :obj:`0.0001`)
        discriminator_lr (float): The learning rate of the discriminator. (defualt: :obj:`0.0001`)
        lambd (float): The weight of gradient penalty loss to trade-off between the adversarial training loss and gradient penalty loss. (defualt: :obj:`1.0`)
        weight_decay: (float): The weight decay (L2 penalty). (defualt: :obj:`0.0`)
        device_ids (list): Use cpu if the list is empty. If the list contains indices of multiple GPUs, it needs to be launched with :obj:`torch.distributed.launch` or :obj:`torchrun`. (defualt: :obj:`[]`)
        ddp_sync_bn (bool): Whether to replace batch normalization in network structure with cross-GPU synchronized batch normalization. Only valid when the length of :obj:`device_ids` is greater than one. (defualt: :obj:`True`)
        ddp_replace_sampler (bool): Whether to replace sampler in dataloader with :obj:`DistributedSampler`. Only valid when the length of :obj:`device_ids` is greater than one. (defualt: :obj:`True`)
        ddp_val (bool): Whether to use multi-GPU acceleration for the validation set. For experiments where data input order is sensitive, :obj:`ddp_val` should be set to :obj:`False`. Only valid when the length of :obj:`device_ids` is greater than one. (defualt: :obj:`True`)
        ddp_test (bool): Whether to use multi-GPU acceleration for the test set. For experiments where data input order is sensitive, :obj:`ddp_test` should be set to :obj:`False`. Only valid when the length of :obj:`device_ids` is greater than one. (defualt: :obj:`True`)
    
    .. automethod:: fit
    .. automethod:: test
    .. automethod:: sample
    '''
    def __init__(self,
                 generator: nn.Module,
                 discriminator: nn.Module,
                 generator_lr: float = 1e-4,
                 discriminator_lr: float = 1e-4,
                 lambd: float = 1.0,
                 weight_decay: float = 0.0,
                 device_ids: List[int] = [],
                 ddp_sync_bn: bool = True,
                 ddp_replace_sampler: bool = True,
                 ddp_val: bool = True,
                 ddp_test: bool = True):
        super(GANTrainer,
              self).__init__(modules={
                  'generator': generator,
                  'discriminator': discriminator,
              },
                             device_ids=device_ids,
                             ddp_sync_bn=ddp_sync_bn,
                             ddp_replace_sampler=ddp_replace_sampler,
                             ddp_val=ddp_val,
                             ddp_test=ddp_test)
        self.generator_lr = generator_lr
        self.discriminator_lr = discriminator_lr
        self.weight_decay = weight_decay
        self.lambd = lambd

        self.generator_optimizer = torch.optim.Adam(generator.parameters(),
                                                    lr=generator_lr,
                                                    weight_decay=weight_decay)
        self.discriminator_optimizer = torch.optim.Adam(
            discriminator.parameters(),
            lr=discriminator_lr,
            weight_decay=weight_decay)

        self.loss_fn = nn.BCEWithLogitsLoss()
        # init metric
        self.train_g_loss = torchmetrics.MeanMetric().to(self.device)
        self.train_d_loss = torchmetrics.MeanMetric().to(self.device)
        self.val_g_loss = torchmetrics.MeanMetric().to(self.device)
        self.val_d_loss = torchmetrics.MeanMetric().to(self.device)
        self.test_g_loss = torchmetrics.MeanMetric().to(self.device)
        self.test_d_loss = torchmetrics.MeanMetric().to(self.device)

    def before_training_epoch(self, epoch_id: int, num_epochs: int):
        self.log(f"Epoch {epoch_id}\n-------------------------------")

    def on_training_step(self, train_batch: Tuple, batch_id: int,
                         num_batches: int):
        self.train_g_loss.reset()
        self.train_d_loss.reset()

        X = train_batch[0].to(self.device)
        y = train_batch[1].to(self.device)

        self.generator_optimizer.zero_grad()

        assert hasattr(
            self.modules['generator'], 'in_channels'
        ), 'The generator must have the property in_channels to generate a batch of latent codes for the corresponding dimension.'

        z = torch.normal(mean=0,
                         std=1,
                         size=(X.shape[0],
                               self.modules['generator'].in_channels)).to(
                                   self.device)
        gen_X = self.modules['generator'](z)
        g_loss = -torch.mean(self.modules['discriminator'](gen_X))

        g_loss.backward()
        self.generator_optimizer.step()

        # backpropagation for discriminator
        self.discriminator_optimizer.zero_grad()

        real_loss = self.modules['discriminator'](X)
        fake_loss = self.modules['discriminator'](gen_X.detach())
        gp_term = gradient_penalty(self.modules['discriminator'], X, gen_X)
        d_loss = -torch.mean(real_loss) + torch.mean(
            fake_loss) + self.lambd * gp_term
        d_loss.backward()
        self.discriminator_optimizer.step()

        # log five times
        log_step = math.ceil(num_batches / 5)
        if batch_id % log_step == 0:
            self.train_g_loss.update(g_loss)
            self.train_d_loss.update(d_loss)

            train_g_loss = self.train_g_loss.compute()
            train_d_loss = self.train_d_loss.compute()

            # if not distributed, world_size is 1
            batch_id = batch_id * self.world_size
            num_batches = num_batches * self.world_size
            if self.is_main:
                self.log(
                    f"g_loss: {train_g_loss:>8f}, d_loss: {train_d_loss:>8f} [{batch_id:>5d}/{num_batches:>5d}]"
                )

    def before_validation_epoch(self, epoch_id: int, num_epochs: int):
        self.val_g_loss.reset()
        self.val_d_loss.reset()

    def on_validation_step(self, val_batch: Tuple, batch_id: int,
                           num_batches: int):
        X = val_batch[0].to(self.device)
        y = val_batch[1].to(self.device)

        # for g_loss
        z = torch.normal(mean=0,
                         std=1,
                         size=(X.shape[0],
                               self.modules['generator'].in_channels)).to(
                                   self.device)
        gen_X = self.modules['generator'](z)
        # g_loss = self.loss_fn(self.modules['discriminator'](gen_X), valid)
        g_loss = -torch.mean(self.modules['discriminator'](gen_X))

        # for d_loss
        real_loss = self.modules['discriminator'](X)
        fake_loss = self.modules['discriminator'](gen_X.detach())
        gp_term = gradient_penalty(self.modules['discriminator'], X, gen_X)
        d_loss = -torch.mean(real_loss) + torch.mean(
            fake_loss) + self.lambd * gp_term

        self.val_g_loss.update(g_loss)
        self.val_d_loss.update(d_loss)

    def after_validation_epoch(self, epoch_id: int, num_epochs: int):
        val_g_loss = self.val_g_loss.compute()
        val_d_loss = self.val_d_loss.compute()
        self.log(f"\ng_loss: {val_g_loss:>8f}, d_loss: {val_d_loss:>8f}")

    def before_test_epoch(self):
        self.test_g_loss.reset()
        self.test_d_loss.reset()

    def on_test_step(self, test_batch: Tuple, batch_id: int, num_batches: int):
        X = test_batch[0].to(self.device)
        y = test_batch[1].to(self.device)

        # for g_loss
        z = torch.normal(mean=0,
                         std=1,
                         size=(X.shape[0],
                               self.modules['generator'].in_channels)).to(
                                   self.device)
        gen_X = self.modules['generator'](z)
        g_loss = -torch.mean(self.modules['discriminator'](gen_X))

        # for d_loss
        real_loss = self.modules['discriminator'](X)
        fake_loss = self.modules['discriminator'](gen_X.detach())
        gp_term = gradient_penalty(self.modules['discriminator'], X, gen_X)
        d_loss = -torch.mean(real_loss) + torch.mean(
            fake_loss) + self.lambd * gp_term

        self.test_g_loss.update(g_loss)
        self.test_d_loss.update(d_loss)

    def after_test_epoch(self):
        test_g_loss = self.test_g_loss.compute()
        test_d_loss = self.test_d_loss.compute()
        self.log(f"\ng_loss: {test_g_loss:>8f}, d_loss: {test_d_loss:>8f}")

    def fit(self,
            train_loader: DataLoader,
            val_loader: DataLoader,
            num_epochs: int = 1,
            **kwargs):
        r'''
        Args:
            train_loader (DataLoader): Iterable DataLoader for traversing the training data batch (torch.utils.data.dataloader.DataLoader, torch_geometric.loader.DataLoader, etc).
            val_loader (DataLoader): Iterable DataLoader for traversing the validation data batch (torch.utils.data.dataloader.DataLoader, torch_geometric.loader.DataLoader, etc).
            num_epochs (int): training epochs. (defualt: :obj:`1`)
        '''
        train_loader = self.on_reveive_dataloader(train_loader, mode='train')
        val_loader = self.on_reveive_dataloader(val_loader, mode='val')

        for t in range(num_epochs):
            if hasattr(train_loader, 'need_to_set_epoch'):
                train_loader.sampler.set_epoch(t)
            if hasattr(val_loader, 'need_to_set_epoch'):
                val_loader.sampler.set_epoch(t)

            num_batches = len(train_loader)

            # set model to train mode
            for k, m in self.modules.items():
                self.modules[k].train()

            # hook
            self.before_training_epoch(t + 1, num_epochs, **kwargs)
            for batch_id, train_batch in enumerate(train_loader):
                # hook
                self.before_training_step(batch_id, num_batches, **kwargs)
                # hook
                self.on_training_step(train_batch, batch_id, num_batches,
                                      **kwargs)
                # hook
                self.after_training_step(batch_id, num_batches, **kwargs)
            # hook
            self.after_training_epoch(t + 1, num_epochs, **kwargs)

            # set model to val mode
            for k, m in self.modules.items():
                self.modules[k].eval()

            num_batches = len(val_loader)
            # hook
            self.before_validation_epoch(t + 1, num_epochs, **kwargs)
            for batch_id, val_batch in enumerate(val_loader):
                # hook
                self.before_validation_step(batch_id, num_batches, **kwargs)
                # hook
                self.on_validation_step(val_batch, batch_id, num_batches,
                                        **kwargs)
                # hook
                self.after_validation_step(batch_id, num_batches, **kwargs)
                # hook
            self.after_validation_epoch(t + 1, num_epochs, **kwargs)

        return self

    def test(self, test_loader: DataLoader, **kwargs):
        r'''
        Args:
            test_loader (DataLoader): Iterable DataLoader for traversing the test data batch (torch.utils.data.dataloader.DataLoader, torch_geometric.loader.DataLoader, etc).
        '''
        test_loader = self.on_reveive_dataloader(test_loader, mode='test')

        for k, m in self.modules.items():
            self.modules[k].eval()

        num_batches = len(test_loader)
        self.before_test_epoch(**kwargs)
        for batch_id, test_batch in enumerate(test_loader):
            # hook
            self.before_test_step(batch_id, num_batches, **kwargs)
            # hook
            self.on_test_step(test_batch, batch_id, num_batches, **kwargs)
            # hook
            self.after_test_step(batch_id, num_batches, **kwargs)
        self.after_test_epoch(**kwargs)

    def sample(self, num_samples: int) -> torch.Tensor:
        """
        Samples from the latent space and return generated results.

        Args:
            num_samples (int): Number of samples.

        Returns:
            torch.Tensor: the generated samples.
        """
        self.modules['generator'].eval()
        with torch.no_grad():
            z = torch.randn(num_samples,
                            self.modules['generator'].in_channels).to(
                                self.device)
            return self.modules['generator'](z)


class CGANTrainer(GANTrainer):
    r'''
    This class provide the implementation for WGAN-GP. It trains a zero-sum game between the generator and the discriminator, just like the traditional generative networks. The generator is optimized to generate simulation samples that are indistinguishable by the discriminator, and the discriminator is optimized to discriminate false samples generated by the generator. Compared with vanilla GAN, with WGAN-GP we can improve the stability of learning, get rid of problems like mode collapse, and provide meaningful learning curves useful for debugging and hyperparameter searches. Thus, existing work typically uses WGAN-GP to generate simulated EEG signals. In particular, the expected labels are additionally provided to guide the discriminator to distinguish whether the sample fits the data distribution of the class. For more details, please refer to the following information. 

    - Paper: Gulrajani I, Ahmed F, Arjovsky M, et al. Improved training of wasserstein gans[J]. Advances in neural information processing systems, 2017, 30.
    - URL: https://arxiv.org/abs/1704.00028
    - Related Project: https://github.com/eriklindernoren/PyTorch-GAN

    .. code-block:: python

        g_model = BCGenerator(in_channels=128, num_classes=2)
        d_model = BCDiscriminator(in_channels=4, num_classes=2)
        trainer = CGANTrainer(generator, discriminator)
        trainer.fit(train_loader, val_loader)
        trainer.test(test_loader)

    The class provides the following hook functions for inserting additional implementations in the training, validation and testing lifecycle:

    - :obj:`before_training_epoch`: executed before each epoch of training starts
    - :obj:`before_training_step`: executed before each batch of training starts
    - :obj:`on_training_step`: the training process for each batch
    - :obj:`after_training_step`: execute after the training of each batch
    - :obj:`after_training_epoch`: executed after each epoch of training
    - :obj:`before_validation_epoch`: executed before each round of validation starts
    - :obj:`before_validation_step`: executed before the validation of each batch
    - :obj:`on_validation_step`: validation process for each batch
    - :obj:`after_validation_step`: executed after the validation of each batch
    - :obj:`after_validation_epoch`: executed after each round of validation
    - :obj:`before_test_epoch`: executed before each round of test starts
    - :obj:`before_test_step`: executed before the test of each batch
    - :obj:`on_test_step`: test process for each batch
    - :obj:`after_test_step`: executed after the test of each batch
    - :obj:`after_test_epoch`: executed after each round of test

    If you want to customize some operations, you just need to inherit the class and override the hook function:

    .. code-block:: python

        class MyGANTrainer(GANTrainer):
            def before_training_epoch(self, epoch_id: int, num_epochs: int):
                # Do something here.
                super().before_training_epoch(epoch_id, num_epochs)
    
    If you want to use multiple GPUs for parallel computing, you need to specify the GPU indices you want to use in the python file:
    
    .. code-block:: python

        trainer = GANTrainer(generator, discriminator, device_ids=[1, 2, 7])
        trainer.fit(train_loader, val_loader)
        trainer.test(test_loader)

    Then, you can use the :obj:`torch.distributed.launch` or :obj:`torchrun` to run your python file.

    .. code-block:: shell

        python -m torch.distributed.launch \
            --nproc_per_node=3 \
            --nnodes=1 \
            --node_rank=0 \
            --master_addr="localhost" \
            --master_port=2345 \
            your_python_file.py

    Here, :obj:`nproc_per_node` is the number of GPUs you specify.

    Args:
        generator (nn.Module): The generator model for EEG signal generation, whose inputs are Gaussian distributed random vectors, outputs are generated EEG signals. The dimensions of the input vector should be defined on the :obj:`in_channel` attribute. The output layer does not need to have a softmax activation function.
        discriminator (nn.Module): The discriminator model to determine whether the EEG signal is real or generated, and the dimension of its output should be equal to the one (i.e., the score to distinguish the real and the fake). The output layer does not need to have a sigmoid activation function.
        generator_lr (float): The learning rate of the generator. (defualt: :obj:`0.0001`)
        discriminator_lr (float): The learning rate of the discriminator. (defualt: :obj:`0.0001`)
        lambd (float): The weight of gradient penalty loss to trade-off between the adversarial training loss and gradient penalty loss. (defualt: :obj:`1.0`)
        weight_decay: (float): The weight decay (L2 penalty). (defualt: :obj:`0.0`)
        device_ids (list): Use cpu if the list is empty. If the list contains indices of multiple GPUs, it needs to be launched with :obj:`torch.distributed.launch` or :obj:`torchrun`. (defualt: :obj:`[]`)
        ddp_sync_bn (bool): Whether to replace batch normalization in network structure with cross-GPU synchronized batch normalization. Only valid when the length of :obj:`device_ids` is greater than one. (defualt: :obj:`True`)
        ddp_replace_sampler (bool): Whether to replace sampler in dataloader with :obj:`DistributedSampler`. Only valid when the length of :obj:`device_ids` is greater than one. (defualt: :obj:`True`)
        ddp_val (bool): Whether to use multi-GPU acceleration for the validation set. For experiments where data input order is sensitive, :obj:`ddp_val` should be set to :obj:`False`. Only valid when the length of :obj:`device_ids` is greater than one. (defualt: :obj:`True`)
        ddp_test (bool): Whether to use multi-GPU acceleration for the test set. For experiments where data input order is sensitive, :obj:`ddp_test` should be set to :obj:`False`. Only valid when the length of :obj:`device_ids` is greater than one. (defualt: :obj:`True`)
    
    .. automethod:: fit
    .. automethod:: test
    .. automethod:: sample
    '''
    def __init__(self,
                 generator: nn.Module,
                 discriminator: nn.Module,
                 generator_lr: float = 1e-4,
                 discriminator_lr: float = 1e-4,
                 lambd: float = 1.0,
                 weight_decay: float = 0.0,
                 device_ids: List[int] = [],
                 ddp_sync_bn: bool = True,
                 ddp_replace_sampler: bool = True,
                 ddp_val: bool = True,
                 ddp_test: bool = True):
        super(CGANTrainer, self).__init__(
            generator=generator,
            discriminator=discriminator,
            generator_lr=generator_lr,
            discriminator_lr=discriminator_lr,
            lambd=lambd,
            weight_decay=weight_decay,
            device_ids=device_ids,
            ddp_sync_bn=ddp_sync_bn,
            ddp_replace_sampler=ddp_replace_sampler,
            ddp_val=ddp_val,
            ddp_test=ddp_test,
        )

    def on_training_step(self, train_batch: Tuple, batch_id: int,
                         num_batches: int):
        self.train_g_loss.reset()
        self.train_d_loss.reset()

        X = train_batch[0].to(self.device)
        y = train_batch[1].to(self.device)

        self.generator_optimizer.zero_grad()

        assert hasattr(
            self.modules['generator'], 'in_channels'
        ), 'The generator must have the property in_channels to generate a batch of latent codes for the corresponding dimension.'

        z = torch.normal(mean=0,
                         std=1,
                         size=(X.shape[0],
                               self.modules['generator'].in_channels)).to(
                                   self.device)
        gen_X = self.modules['generator'](z, y)
        g_loss = -torch.mean(self.modules['discriminator'](gen_X, y))

        g_loss.backward()
        self.generator_optimizer.step()

        # backpropagation for discriminator
        self.discriminator_optimizer.zero_grad()

        real_loss = self.modules['discriminator'](X, y)
        fake_loss = self.modules['discriminator'](gen_X.detach(), y)
        gp_term = gradient_penalty(self.modules['discriminator'], X, gen_X, y)
        d_loss = -torch.mean(real_loss) + torch.mean(
            fake_loss) + self.lambd * gp_term

        d_loss.backward()
        self.discriminator_optimizer.step()

        # log five times
        log_step = math.ceil(num_batches / 5)
        if batch_id % log_step == 0:
            self.train_g_loss.update(g_loss)
            self.train_d_loss.update(d_loss)

            train_g_loss = self.train_g_loss.compute()
            train_d_loss = self.train_d_loss.compute()

            # if not distributed, world_size is 1
            batch_id = batch_id * self.world_size
            num_batches = num_batches * self.world_size
            if self.is_main:
                self.log(
                    f"g_loss: {train_g_loss:>8f}, d_loss: {train_d_loss:>8f} [{batch_id:>5d}/{num_batches:>5d}]"
                )

    def on_validation_step(self, val_batch: Tuple, batch_id: int,
                           num_batches: int):
        X = val_batch[0].to(self.device)
        y = val_batch[1].to(self.device)

        # for g_loss
        z = torch.normal(mean=0,
                         std=1,
                         size=(X.shape[0],
                               self.modules['generator'].in_channels)).to(
                                   self.device)
        gen_X = self.modules['generator'](z, y)
        g_loss = -torch.mean(self.modules['discriminator'](gen_X, y))

        # for d_loss
        real_loss = self.modules['discriminator'](X, y)
        fake_loss = self.modules['discriminator'](gen_X.detach(), y)
        gp_term = gradient_penalty(self.modules['discriminator'], X, gen_X, y)
        d_loss = -torch.mean(real_loss) + torch.mean(
            fake_loss) + self.lambd * gp_term

        self.val_g_loss.update(g_loss)
        self.val_d_loss.update(d_loss)

    def on_test_step(self, test_batch: Tuple, batch_id: int, num_batches: int):
        X = test_batch[0].to(self.device)
        y = test_batch[1].to(self.device)

        # for g_loss
        z = torch.normal(mean=0,
                         std=1,
                         size=(X.shape[0],
                               self.modules['generator'].in_channels)).to(
                                   self.device)
        gen_X = self.modules['generator'](z, y)
        g_loss = -torch.mean(self.modules['discriminator'](gen_X, y))

        # for d_loss
        real_loss = self.modules['discriminator'](X, y)
        fake_loss = self.modules['discriminator'](gen_X.detach(), y)
        gp_term = gradient_penalty(self.modules['discriminator'], X, gen_X, y)
        d_loss = -torch.mean(real_loss) + torch.mean(
            fake_loss) + self.lambd * gp_term

        self.test_g_loss.update(g_loss)
        self.test_d_loss.update(d_loss)

    def sample(self, num_samples: int, labels: torch.Tensor = None) -> torch.Tensor:
        """
        Samples from the latent space and return generated results.

        Args:
            num_samples (int): Number of samples.
            labels (torch.Tensor): Category labels (int) for a batch of samples The shape should be :obj:`[n,]`. Here, :obj:`n` corresponds to the batch size. If not provided, a batch of randomly generated categories will be used.

        Returns:
            torch.Tensor: the generated samples.
        """
        if labels:
            assert len(
                labels
            ) == num_samples, f'labels ({len(labels)}) should be the same length as num_samples ({num_samples}).'
            assert isinstance(
                labels, torch.Tensor
            ), f'labels should be torch.Tensor instances, the current input is {type(labels)}'
        else:
            labels = torch.randint(low=0,
                                   high=self.modules['generator'].num_classes,
                                   size=(num_samples, ))
        labels = labels.long().to(self.device)
        self.modules['generator'].eval()
        with torch.no_grad():
            z = torch.randn(num_samples,
                            self.modules['generator'].in_channels).to(
                                self.device)
            return self.modules['generator'](z, labels)



import math
from typing import List, Tuple

import torch
import numpy as np
import torch.nn as nn
import torchmetrics
from torch.utils.data import DataLoader

from .basic_trainer import BasicTrainer


class ClassificationTrainer(BasicTrainer):
    r'''
    A generic trainer class for EEG classification.

    .. code-block:: python

        trainer = ClassificationTrainer(model)
        trainer.fit(train_loader, val_loader)
        trainer.test(test_loader)

    The class provides the following hook functions for inserting additional implementations in the training, validation and testing lifecycle:

    - :obj:`before_training_epoch`: executed before each epoch of training starts
    - :obj:`before_training_step`: executed before each batch of training starts
    - :obj:`on_training_step`: the training process for each batch
    - :obj:`after_training_step`: execute after the training of each batch
    - :obj:`after_training_epoch`: executed after each epoch of training
    - :obj:`before_validation_epoch`: executed before each round of validation starts
    - :obj:`before_validation_step`: executed before the validation of each batch
    - :obj:`on_validation_step`: validation process for each batch
    - :obj:`after_validation_step`: executed after the validation of each batch
    - :obj:`after_validation_epoch`: executed after each round of validation
    - :obj:`before_test_epoch`: executed before each round of test starts
    - :obj:`before_test_step`: executed before the test of each batch
    - :obj:`on_test_step`: test process for each batch
    - :obj:`after_test_step`: executed after the test of each batch
    - :obj:`after_test_epoch`: executed after each round of test

    If you want to customize some operations, you just need to inherit the class and override the hook function:

    .. code-block:: python

        class MyClassificationTrainer(ClassificationTrainer):
            def before_training_epoch(self, epoch_id: int, num_epochs: int):
                # Do something here.
                super().before_training_epoch(epoch_id, num_epochs)
    
    If you want to use multiple GPUs for parallel computing, you need to specify the GPU indices you want to use in the python file:
    
    .. code-block:: python

        trainer = ClassificationTrainer(model, device_ids=[1, 2, 7])
        trainer.fit(train_loader, val_loader)
        trainer.test(test_loader)

    Then, you can use the :obj:`torch.distributed.launch` or :obj:`torchrun` to run your python file.

    .. code-block:: shell

        python -m torch.distributed.launch \
            --nproc_per_node=3 \
            --nnodes=1 \
            --node_rank=0 \
            --master_addr="localhost" \
            --master_port=2345 \
            your_python_file.py

    Here, :obj:`nproc_per_node` is the number of GPUs you specify.

    Args:
        model (nn.Module): The classification model, and the dimension of its output should be equal to the number of categories in the dataset. The output layer does not need to have a softmax activation function.
        lr (float): The learning rate. (defualt: :obj:`0.0001`)
        weight_decay: (float): The weight decay (L2 penalty). (defualt: :obj:`0.0`)
        device_ids (list): Use cpu if the list is empty. If the list contains indices of multiple GPUs, it needs to be launched with :obj:`torch.distributed.launch` or :obj:`torchrun`. (defualt: :obj:`[]`)
        ddp_sync_bn (bool): Whether to replace batch normalization in network structure with cross-GPU synchronized batch normalization. Only valid when the length of :obj:`device_ids` is greater than one. (defualt: :obj:`True`)
        ddp_replace_sampler (bool): Whether to replace sampler in dataloader with :obj:`DistributedSampler`. Only valid when the length of :obj:`device_ids` is greater than one. (defualt: :obj:`True`)
        ddp_val (bool): Whether to use multi-GPU acceleration for the validation set. For experiments where data input order is sensitive, :obj:`ddp_val` should be set to :obj:`False`. Only valid when the length of :obj:`device_ids` is greater than one. (defualt: :obj:`True`)
        ddp_test (bool): Whether to use multi-GPU acceleration for the test set. For experiments where data input order is sensitive, :obj:`ddp_test` should be set to :obj:`False`. Only valid when the length of :obj:`device_ids` is greater than one. (defualt: :obj:`True`)
    
    .. automethod:: fit
    .. automethod:: test
    '''
    def __init__(self,
                 model: nn.Module,
                 lr: float = 1e-4,
                 weight_decay: float = 0.0,
                 device_ids: List[int] = [],
                 ddp_sync_bn: bool = True,
                 ddp_replace_sampler: bool = True,
                 ddp_val: bool = True,
                 ddp_test: bool = True):
        super(ClassificationTrainer,
              self).__init__(modules={'model': model},
                             device_ids=device_ids,
                             ddp_sync_bn=ddp_sync_bn,
                             ddp_replace_sampler=ddp_replace_sampler,
                             ddp_val=ddp_val,
                             ddp_test=ddp_test)
        self.lr = lr
        self.weight_decay = weight_decay

        self.optimizer = torch.optim.Adam(model.parameters(),
                                          lr=lr,
                                          weight_decay=weight_decay)
        self.loss_fn = nn.CrossEntropyLoss()

        # init metric
        self.train_loss = torchmetrics.MeanMetric().to(self.device)
        self.train_accuracy = torchmetrics.Accuracy().to(self.device)

        self.val_loss = torchmetrics.MeanMetric().to(self.device)
        self.val_accuracy = torchmetrics.Accuracy().to(self.device)

        self.test_loss = torchmetrics.MeanMetric().to(self.device)
        self.test_accuracy = torchmetrics.Accuracy().to(self.device)

    def before_training_epoch(self, epoch_id: int, num_epochs: int, **kwargs):
        self.log(f"Epoch {epoch_id}\n-------------------------------")

    def on_training_step(self, train_batch: Tuple, batch_id: int,
                         num_batches: int, **kwargs):
        self.train_accuracy.reset()
        self.train_loss.reset()

        X = train_batch[0].to(self.device)
        y = train_batch[1].to(self.device)

        # compute prediction error
        pred = self.modules['model'](X)
        loss = self.loss_fn(pred, y)

        # backpropagation
        self.optimizer.zero_grad()
        loss.backward()
        self.optimizer.step()

        # log five times
        log_step = math.ceil(num_batches / 5)
        if batch_id % log_step == 0:
            self.train_loss.update(loss)
            self.train_accuracy.update(pred.argmax(1), y)

            train_loss = self.train_loss.compute()
            train_accuracy = 100 * self.train_accuracy.compute()

            # if not distributed, world_size is 1
            batch_id = batch_id * self.world_size
            num_batches = num_batches * self.world_size
            if self.is_main:
                self.log(
                    f"loss: {train_loss:>8f}, accuracy: {train_accuracy:>0.1f}% [{batch_id:>5d}/{num_batches:>5d}]"
                )

    def before_validation_epoch(self, epoch_id: int, num_epochs: int, **kwargs):
        self.val_accuracy.reset()
        self.val_loss.reset()

    def on_validation_step(self, val_batch: Tuple, batch_id: int,
                           num_batches: int, **kwargs):
        X = val_batch[0].to(self.device)
        y = val_batch[1].to(self.device)

        pred = self.modules['model'](X)

        self.val_loss.update(self.loss_fn(pred, y))
        self.val_accuracy.update(pred.argmax(1), y)

    def after_validation_epoch(self, epoch_id: int, num_epochs: int, **kwargs):
        val_accuracy = 100 * self.val_accuracy.compute()
        val_loss = self.val_loss.compute()
        self.log(f"\nloss: {val_loss:>8f}, accuracy: {val_accuracy:>0.1f}%")

    def before_test_epoch(self, **kwargs):
        self.test_loss.reset()
        self.test_accuracy.reset()

    def on_test_step(self, test_batch: Tuple, batch_id: int, num_batches: int,
                     **kwargs):
        X = test_batch[0].to(self.device)
        y = test_batch[1].to(self.device)
        pred = self.modules['model'](X)

        self.test_loss.update(self.loss_fn(pred, y))
        self.test_accuracy.update(pred.argmax(1), y)

    def after_test_epoch(self, **kwargs):
        test_accuracy = 100 * self.test_accuracy.compute()
        test_loss = self.test_loss.compute()
        self.log(f"\nloss: {test_loss:>8f}, accuracy: {test_accuracy:>0.1f}%")

    def test(self, test_loader: DataLoader, **kwargs):
        r'''
        Args:
            test_loader (DataLoader): Iterable DataLoader for traversing the test data batch (torch.utils.data.dataloader.DataLoader, torch_geometric.loader.DataLoader, etc).
        '''
        super().test(test_loader=test_loader, **kwargs)

    def fit(self,
            train_loader: DataLoader,
            val_loader: DataLoader,
            num_epochs: int = 1,
            **kwargs):
        r'''
        Args:
            train_loader (DataLoader): Iterable DataLoader for traversing the training data batch (torch.utils.data.dataloader.DataLoader, torch_geometric.loader.DataLoader, etc).
            val_loader (DataLoader): Iterable DataLoader for traversing the validation data batch (torch.utils.data.dataloader.DataLoader, torch_geometric.loader.DataLoader, etc).
            num_epochs (int): training epochs. (defualt: :obj:`1`)
        '''
        super().fit(train_loader=train_loader,
                    val_loader=val_loader,
                    num_epochs=num_epochs,
                    **kwargs)



import os
from functools import partial
from multiprocessing import Manager, Pool, Process, Queue
from typing import Callable, Dict, Tuple, Union

from torcheeg.io import EEGSignalIO, MetaInfoIO
from tqdm import tqdm

import numpy as np

MAX_QUEUE_SIZE = 1024


def transform_producer(X_y_rank: Tuple[np.ndarray, Dict,
                                       int], before_trial: Union[Callable,
                                                                 None],
                       transform: Union[Callable,
                                        None], after_trial: Union[Callable,
                                                                  None],
                       write_info_fn: Callable, queue: Queue):
    X, y, worker_rank = X_y_rank
    if before_trial:
        X = before_trial(X)

    trial_queue = []
    for write_pointer, clip_sample in enumerate(X):

        t_eeg = clip_sample

        if not transform is None:
            t = transform(eeg=clip_sample)
            t_eeg = t['eeg']

        clip_id = f'{worker_rank}_{write_pointer}'
        if after_trial:
            trial_queue.append({'eeg': t_eeg, 'key': clip_id})
        else:
            queue.put({'eeg': t_eeg, 'key': clip_id})

        # record meta info for each signal
        record_info = {'clip_id': clip_id}
        record_info.update({k: v[write_pointer] for k, v in y.items()})
        write_info_fn(record_info)

    if len(trial_queue) and after_trial:
        trial_queue = after_trial(trial_queue)
        for obj in trial_queue:
            assert 'eeg' in obj and 'key' in obj, 'after_trial must return a list of dictionaries, where each dictionary corresponds to an EEG sample, containing `eeg` and `key` as keys.'
            queue.put(obj)


def io_consumer(write_eeg_fn, queue):
    while True:
        item = queue.get()
        if not item is None:
            eeg = item['eeg']
            key = item['key']
            write_eeg_fn(eeg, key)
        else:
            break


class SingleProcessingQueue:
    def __init__(self, write_eeg_fn):
        self.write_eeg_fn = write_eeg_fn

    def put(self, item):
        eeg = item['eeg']
        key = item['key']
        self.write_eeg_fn(eeg, key)


def numpy_constructor(X: np.ndarray,
                      y: Dict,
                      before_trial: Union[None, Callable] = None,
                      transform: Union[None, Callable] = None,
                      after_trial: Union[Callable, None] = None,
                      io_path: str = './io/numpy',
                      num_worker: int = 0,
                      num_samples_per_trial: int = 100,
                      verbose: bool = True,
                      cache_size: int = 64 * 1024 * 1024 * 1024) -> None:

    for k, v in y.items():
        assert len(X) == len(
            v
        ), f'The number of labels {len(v)} does not match the number of samples {len(X)} for labels {k}.'

    # init IO
    meta_info_io_path = os.path.join(io_path, 'info.csv')
    eeg_signal_io_path = os.path.join(io_path, 'eeg')

    if os.path.exists(io_path) and not os.path.getsize(meta_info_io_path) == 0:
        print(
            f'The target folder already exists, if you need to regenerate the database IO, please delete the path {io_path}.'
        )
        return

    os.makedirs(io_path, exist_ok=True)

    info_io = MetaInfoIO(meta_info_io_path)
    eeg_io = EEGSignalIO(eeg_signal_io_path, cache_size=cache_size)

    # access the data
    indices = np.arange(len(X))
    worker_indices_list = np.array_split(indices,
                                         len(X) // num_samples_per_trial)

    X_y_rank_list = []
    for worker_rank, worker_indices in enumerate(worker_indices_list):
        X_worker = X[worker_indices]
        y_worker = {k: v[worker_indices] for k, v in y.items()}
        X_y_rank_list.append((X_worker, y_worker, worker_rank))

    if verbose:
        # show process bar
        pbar = tqdm(total=len(X_y_rank_list))
        pbar.set_description("[NUMPY]")

    if num_worker > 1:
        manager = Manager()
        queue = manager.Queue(maxsize=MAX_QUEUE_SIZE)
        io_consumer_process = Process(target=io_consumer,
                                      args=(eeg_io.write_eeg, queue),
                                      daemon=True)
        io_consumer_process.start()

        partial_mp_fn = partial(transform_producer,
                                before_trial=before_trial,
                                transform=transform,
                                after_trial=after_trial,
                                write_info_fn=info_io.write_info,
                                queue=queue)

        for _ in Pool(num_worker).imap(partial_mp_fn, X_y_rank_list):
            if verbose:
                pbar.update(1)

        queue.put(None)

        io_consumer_process.join()
        io_consumer_process.close()

    else:
        for X_y_rank in X_y_rank_list:
            transform_producer(X_y_rank=X_y_rank,
                               before_trial=before_trial,
                               transform=transform,
                               after_trial=after_trial,
                               write_info_fn=info_io.write_info,
                               queue=SingleProcessingQueue(eeg_io.write_eeg))
            if verbose:
                pbar.update(1)

    if verbose:
        pbar.close()
        print('Please wait for the writing process to complete...')



from typing import Callable, Dict, Tuple, Union

from ...constants.ssvep.tsu_benchmark import (TSUBENCHMARK_ADJACENCY_MATRIX,
                                                   TSUBENCHMARK_CHANNEL_LOCATION_DICT)
from ...functional.ssvep.tsu_benchmark import tsu_benchmark_constructor
from ..base_dataset import BaseDataset


class TSUBenckmarkDataset(BaseDataset):
    r'''
    The benchmark dataset for SSVEP-Based brain-computer interfaces (TSUBenckmark) is provided by the Tsinghua BCI Lab. It presents a benchmark steady-state visual evoked potential (SSVEP) dataset acquired with a 40-target brain-computer interface (BCI) speller. This class generates training samples and test samples according to the given parameters, and caches the generated results in a unified input and output format (IO). The relevant information of the dataset is as follows:

    - Author: Wang et al.
    - Year: 2016
    - Download URL: http://bci.med.tsinghua.edu.cn/
    - Reference: Wang Y, Chen X, Gao X, et al. A benchmark dataset for SSVEP-based brain-computer interfaces[J]. IEEE Transactions on Neural Systems and Rehabilitation Engineering, 2016, 25(10): 1746-1752.
    - Stimulus: Each trial started with a visual cue (a red square) indicating a target stimulus. The cue appeared for 0.5s on the screen. Subjects were asked to shift their gaze to the target as soon as possible within the cue duration. Following the cue offset, all stimuli started to flicker on the screen concurrently and lasted 5s. After stimulus offset, the screen was blank for 0.5s before the next trial began, which allowed the subjects to have short breaks between consecutive trials.
    - Signals: Electroencephalogram (64 channels at 250Hz) of 35 subjects. For each subject, the experiment consisted of 6 blocks. Each block contained 40 trials corresponding to all 40 characters indicated in a random order. Totally 35 people x 6 blocks x 40 trials.
    - Rating: Frequency and phase values for the 40 trials.

    In order to use this dataset, the download folder :obj:`data_preprocessed_python` is required, containing the following files:
    
    - Readme.txt
    - Sub_info.txt
    - 64-channels.loc
    - Freq_Phase.mat
    - S1.mat
    - ...
    - S35.mat

    An example dataset for CNN-based methods:

    .. code-block:: python

        dataset = TSUBenckmarkDataset(io_path=f'./tsu_benchmark',
                              root_path='./TSUBenchmark',
                              offline_transform=transforms.Compose([
                                  transforms.BandDifferentialEntropy(),
                                  transforms.ToGrid(TSUBenckmark_CHANNEL_LOCATION_DICT)
                              ]),
                              online_transform=transforms.ToTensor(),
                              label_transform=transforms.Select(['trial_id']))
        print(dataset[0])
        # EEG signal (torch.Tensor[250, 10, 11]),
        # coresponding baseline signal (torch.Tensor[250, 10, 11]),
        # label (int)

    Another example dataset for CNN-based methods:

    .. code-block:: python

        dataset = TSUBenckmarkDataset(io_path=f'./tsu_benchmark',
                              root_path='./TSUBenchmark',
                              online_transform=transforms.Compose([
                                  transforms.ToTensor(),
                                  transforms.To2d()
                              ]),
                              label_transform=transforms.Select(['trial_id']))
        print(dataset[0])
        # EEG signal (torch.Tensor[64, 250]),
        # coresponding baseline signal (torch.Tensor[64, 250]),
        # label (int)

    An example dataset for GNN-based methods:

    .. code-block:: python
    
        dataset = TSUBenckmarkDataset(io_path=f'./tsu_benchmark',
                              root_path='./TSUBenchmark',
                              online_transform=transforms.Compose([
                                  transforms.pyg.ToG(TSUBenckmark_ADJACENCY_MATRIX)
                              ]),
                              label_transform=transforms.Select(['trial_id']))
        print(dataset[0])
        # EEG signal (torch_geometric.data.Data),
        # coresponding baseline signal (torch_geometric.data.Data),
        # label (int)
        
    In particular, TorchEEG utilizes the producer-consumer model to allow multi-process data preprocessing. If your data preprocessing is time consuming, consider increasing :obj:`num_worker` for higher speedup. If running under Windows, please use the proper idiom in the main module:

    .. code-block:: python
    
        if __name__ == '__main__':
            dataset = TSUBenckmarkDataset(io_path=f'./tsu_benchmark',
                              root_path='./TSUBenchmark',
                              online_transform=transforms.Compose([
                                  transforms.pyg.ToG(TSUBenckmark_ADJACENCY_MATRIX)
                              ]),
                              label_transform=transforms.Select(['freq']),
                              num_worker=4)
            print(dataset[0])
            # EEG signal (torch_geometric.data.Data),
            # coresponding baseline signal (torch_geometric.data.Data),
            # label (int)

    Args:
        root_path (str): Downloaded data files in matlab (unzipped TSUBenchmark.zip) formats (default: :obj:`'./TSUBenchmark'`)
        chunk_size (int): Number of data points included in each EEG chunk as training or test samples. (default: :obj:`250`)
        overlap (int): The number of overlapping data points between different chunks when dividing EEG chunks. (default: :obj:`0`)
        num_channel (int): Number of channels used, of which the first 64 channels are EEG signals. (default: :obj:`64`)
        online_transform (Callable, optional): The transformation of the EEG signals and baseline EEG signals. The input is a :obj:`np.ndarray`, and the ouput is used as the first and second value of each element in the dataset. (default: :obj:`None`)
        offline_transform (Callable, optional): The usage is the same as :obj:`online_transform`, but executed before generating IO intermediate results. (default: :obj:`None`)
        label_transform (Callable, optional): The transformation of the label. The input is an information dictionary, and the ouput is used as the third value of each element in the dataset. (default: :obj:`None`)
        io_path (str): The path to generated unified data IO, cached as an intermediate result. (default: :obj:`./io/tsu_benchmark`)
        num_worker (str): How many subprocesses to use for data processing. (default: :obj:`0`)
        verbose (bool): Whether to display logs during processing, such as progress bars, etc. (default: :obj:`True`)
    
    '''
    channel_location_dict = TSUBENCHMARK_CHANNEL_LOCATION_DICT
    adjacency_matrix = TSUBENCHMARK_ADJACENCY_MATRIX

    def __init__(self,
                 root_path: str = './TSUBenchmark',
                 chunk_size: int = 250,
                 overlap: int = 0,
                 num_channel: int = 64,
                 online_transform: Union[None, Callable] = None,
                 offline_transform: Union[None, Callable] = None,
                 label_transform: Union[None, Callable] = None,
                 io_path: str = './io/tsu_benchmark',
                 num_worker: int = 0,
                 verbose: bool = True,
                 cache_size: int = 64 * 1024 * 1024 * 1024):
        tsu_benchmark_constructor(root_path=root_path,
                         chunk_size=chunk_size,
                         overlap=overlap,
                         num_channel=num_channel,
                         transform=offline_transform,
                         io_path=io_path,
                         num_worker=num_worker,
                         verbose=verbose,
                         cache_size=cache_size)
        super().__init__(io_path)

        self.root_path = root_path
        self.chunk_size = chunk_size
        self.overlap = overlap
        self.num_channel = num_channel
        self.online_transform = online_transform
        self.offline_transform = offline_transform
        self.label_transform = label_transform
        self.num_worker = num_worker
        self.verbose = verbose
        self.cache_size = cache_size

    def __getitem__(self, index: int) -> Tuple[any, any, int, int, int]:
        info = self.info.iloc[index].to_dict()

        eeg_index = str(info['clip_id'])
        eeg = self.eeg_io.read_eeg(eeg_index)

        signal = eeg
        label = info

        if self.online_transform:
            signal = self.online_transform(eeg=eeg)['eeg']

        if self.label_transform:
            label = self.label_transform(y=info)['y']

        return signal, label

    @property
    def repr_body(self) -> Dict:
        return dict(
            super().repr_body, **{
                'root_path': self.root_path,
                'chunk_size': self.chunk_size,
                'overlap': self.overlap,
                'num_channel': self.num_channel,
                'online_transform': self.online_transform,
                'offline_transform': self.offline_transform,
                'label_transform': self.label_transform,
                'num_worker': self.num_worker,
                'verbose': self.verbose,
                'cache_size': self.cache_size
            })



import os
import re
from functools import partial
from multiprocessing import Manager, Pool, Process, Queue, set_start_method
from typing import Callable, List, Union

import scipy.io as scio
from torcheeg.io import EEGSignalIO, MetaInfoIO
from tqdm import tqdm

MAX_QUEUE_SIZE = 1024


def transform_producer(file_name: str, root_path: str, chunk_size: int,
                       overlap: int, num_channel: int,
                       before_trial: Union[Callable, None],
                       transform: Union[Callable,
                                        None], after_trial: Union[Callable,
                                                                  None],
                       write_info_fn: Callable, queue: Queue):
    subject = int(re.findall(r'S(\d*).mat', file_name)[0])  # subject (35)
    freq_phase = scio.loadmat(os.path.join(root_path, 'Freq_Phase.mat'))
    freqs = freq_phase['freqs'][0]
    phases = freq_phase['phases'][0]

    samples = scio.loadmat(os.path.join(root_path,
                                        file_name))['data'].transpose(
                                            2, 3, 0, 1)
    # 40, 6, 64, 1500
    # Target number: 40
    # Block number: 6
    # Electrode number: 64
    # Time points: 1500

    write_pointer = 0

    for trial_id in range(samples.shape[0]):
        trial_meta_info = {
            'subject_id': subject,
            'trial_id': trial_id,
            'phases': phases[trial_id],
            'freqs': freqs[trial_id]
        }
        trial_samples = samples[trial_id]

        for block_id in range(trial_samples.shape[0]):
            block_meta_info = {'block_id': block_id}
            block_meta_info.update(trial_meta_info)
            block_samples = trial_samples[block_id]
            if before_trial:
                block_samples = before_trial(block_samples)

            start_at = 0
            if chunk_size <= 0:
                chunk_size = block_samples.shape[1] - start_at

            # chunk with chunk size
            end_at = chunk_size
            # calculate moving step
            step = chunk_size - overlap

            block_queue = []
            while end_at <= block_samples.shape[1]:
                clip_sample = block_samples[:num_channel, start_at:end_at]

                t_eeg = clip_sample
                if not transform is None:
                    t_eeg = transform(eeg=clip_sample)['eeg']

                clip_id = f'{file_name}_{write_pointer}'
                if after_trial:
                    block_queue.append({'eeg': t_eeg, 'key': clip_id})
                else:
                    queue.put({'eeg': t_eeg, 'key': clip_id})
                write_pointer += 1

                # record meta info for each signal
                record_info = {
                    'start_at': start_at,
                    'end_at': end_at,
                    'clip_id': clip_id
                }
                record_info.update(block_meta_info)
                write_info_fn(record_info)

                start_at = start_at + step
                end_at = start_at + chunk_size

            if len(block_queue) and after_trial:
                block_queue = after_trial(block_queue)
                for obj in block_queue:
                    assert 'eeg' in obj and 'key' in obj, 'after_trial must return a list of dictionaries, where each dictionary corresponds to an EEG sample, containing `eeg` and `key` as keys.'
                    queue.put(obj)


def io_consumer(write_eeg_fn, queue):
    while True:
        item = queue.get()
        if not item is None:
            eeg = item['eeg']
            key = item['key']
            write_eeg_fn(eeg, key)
        else:
            break


class SingleProcessingQueue:
    def __init__(self, write_eeg_fn):
        self.write_eeg_fn = write_eeg_fn

    def put(self, item):
        eeg = item['eeg']
        key = item['key']
        self.write_eeg_fn(eeg, key)


def tsu_benchmark_constructor(
        root_path: str = './TSUBenchmark',
        chunk_size: int = 200,
        overlap: int = 0,
        num_channel: int = 64,
        before_trial: Union[None, Callable] = None,
        transform: Union[None, Callable] = None,
        after_trial: Union[Callable, None] = None,
        io_path: str = './io/tsu_benchmark',
        num_worker: int = 0,
        verbose: bool = True,
        cache_size: int = 64 * 1024 * 1024 * 1024) -> None:
    # init IO
    meta_info_io_path = os.path.join(io_path, 'info.csv')
    eeg_signal_io_path = os.path.join(io_path, 'eeg')

    if os.path.exists(io_path) and not os.path.getsize(meta_info_io_path) == 0:
        print(
            f'The target folder already exists, if you need to regenerate the database IO, please delete the path {io_path}.'
        )
        return

    os.makedirs(io_path, exist_ok=True)

    meta_info_io_path = os.path.join(io_path, 'info.csv')
    eeg_signal_io_path = os.path.join(io_path, 'eeg')

    info_io = MetaInfoIO(meta_info_io_path)
    eeg_io = EEGSignalIO(eeg_signal_io_path, cache_size=cache_size)

    # loop to access the dataset files
    file_list = os.listdir(root_path)
    skip_set = [
        'Readme.txt', 'Sub_info.txt', '64-channels.loc', '64-channels.loc',
        'Freq_Phase.mat'
    ]
    file_list = [f for f in file_list if f not in skip_set]

    if verbose:
        # show process bar
        pbar = tqdm(total=len(file_list))
        pbar.set_description("[TSU Benchmark]")

    if num_worker > 1:
        manager = Manager()
        queue = manager.Queue(maxsize=MAX_QUEUE_SIZE)
        io_consumer_process = Process(target=io_consumer,
                                      args=(eeg_io.write_eeg, queue),
                                      daemon=True)
        io_consumer_process.start()

        partial_mp_fn = partial(transform_producer,
                                root_path=root_path,
                                chunk_size=chunk_size,
                                overlap=overlap,
                                num_channel=num_channel,
                                before_trial=before_trial,
                                transform=transform,
                                after_trial=after_trial,
                                write_info_fn=info_io.write_info,
                                queue=queue)

        for _ in Pool(num_worker).imap(partial_mp_fn, file_list):
            if verbose:
                pbar.update(1)

        queue.put(None)

        io_consumer_process.join()
        io_consumer_process.close()

    else:
        for file_name in file_list:
            transform_producer(file_name=file_name,
                               root_path=root_path,
                               chunk_size=chunk_size,
                               overlap=overlap,
                               num_channel=num_channel,
                               before_trial=before_trial,
                               transform=transform,
                               after_trial=after_trial,
                               write_info_fn=info_io.write_info,
                               queue=SingleProcessingQueue(eeg_io.write_eeg))
            if verbose:
                pbar.update(1)

    if verbose:
        pbar.close()
        print('Please wait for the writing process to complete...')



import os
from typing import List, Tuple, Dict

import torch
import torch.nn as nn
import torch.distributed as dist

from torch.nn.parallel import DistributedDataParallel
from torch.utils.data import DataLoader, DistributedSampler, SequentialSampler


class BasicTrainer:
    r'''
    A generic trainer class for EEG analysis providing interfaces for all trainers to implement contexts common in training deep learning models. After inheriting this class, :obj:`on_training_step`, :obj:`on_validation_step`, and :obj:`on_test_step` must be implemented.

    The class provides the following hook functions for inserting additional implementations in the training, validation and testing lifecycle:
    
    - :obj:`before_training_epoch`: executed before each epoch of training starts
    - :obj:`before_training_step`: executed before each batch of training starts
    - :obj:`on_training_step`: the training process for each batch
    - :obj:`after_training_step`: execute after the training of each batch
    - :obj:`after_training_epoch`: executed after each epoch of training
    - :obj:`before_validation_epoch`: executed before each round of validation starts
    - :obj:`before_validation_step`: executed before the validation of each batch
    - :obj:`on_validation_step`: validation process for each batch
    - :obj:`after_validation_step`: executed after the validation of each batch
    - :obj:`after_validation_epoch`: executed after each round of validation
    - :obj:`before_test_epoch`: executed before each round of test starts
    - :obj:`before_test_step`: executed before the test of each batch
    - :obj:`on_test_step`: test process for each batch
    - :obj:`after_test_step`: executed after the test of each batch
    - :obj:`after_test_epoch`: executed after each round of test

    If you want to use multiple GPUs for parallel computing, you need to specify the GPU indices you want to use in the python file:
    
    .. code-block:: python

        trainer = BasicTrainer(model, device_ids=[1, 2, 7])
        trainer.fit(train_loader, val_loader)
        trainer.test(test_loader)

    Then, you can use the :obj:`torch.distributed.launch` or :obj:`torchrun` to run your python file.

    .. code-block:: shell

        python -m torch.distributed.launch \
            --nproc_per_node=3 \
            --nnodes=1 \
            --node_rank=0 \
            --master_addr="localhost" \
            --master_port=2345 \
            your_python_file.py

    Here, :obj:`nproc_per_node` is the number of GPUs you specify.

    Args:
        model (Dict): A dictionary that stores neural networks for import, export and device conversion of neural networks.
        device_ids (list): Use cpu if the list is empty. If the list contains indices of multiple GPUs, it needs to be launched with :obj:`torch.distributed.launch` or :obj:`torchrun`. (defualt: :obj:`[]`)
        ddp_sync_bn (bool): Whether to replace batch normalization in network structure with cross-GPU synchronized batch normalization. Only valid when the length of :obj:`device_ids` is greater than one. (defualt: :obj:`True`)
        ddp_replace_sampler (bool): Whether to replace sampler in dataloader with :obj:`DistributedSampler`. Only valid when the length of :obj:`device_ids` is greater than one. (defualt: :obj:`True`)
        ddp_val (bool): Whether to use multi-GPU acceleration for the validation set. For experiments where data input order is sensitive, :obj:`ddp_val` should be set to :obj:`False`. Only valid when the length of :obj:`device_ids` is greater than one. (defualt: :obj:`True`)
        ddp_test (bool): Whether to use multi-GPU acceleration for the test set. For experiments where data input order is sensitive, :obj:`ddp_test` should be set to :obj:`False`. Only valid when the length of :obj:`device_ids` is greater than one. (defualt: :obj:`True`)
    
    .. automethod:: fit
    .. automethod:: test
    '''
    def __init__(self,
                 modules: Dict,
                 device_ids: List[int] = [],
                 ddp_sync_bn: bool = True,
                 ddp_replace_sampler: bool = True,
                 ddp_val: bool = True,
                 ddp_test: bool = True):
        # given params
        self.modules = modules
        self.device_ids = device_ids
        self.ddp_sync_bn = ddp_sync_bn
        self.ddp_replace_sampler = ddp_replace_sampler
        self.ddp_val = ddp_val
        self.ddp_test = ddp_test

        # built-in params
        self.rank = -1
        self.local_rank = -1
        self.world_size = 1
        self.device = None

        # cpu trainer
        if len(device_ids) == 0:
            for k, m in self.modules.items():
                self.modules[k] = m.to(torch.device('cpu'))
            self.device = torch.device('cpu')

        # gpu trainer
        if len(device_ids) == 1:
            assert torch.cuda.is_available(
            ), 'GPU is not available, please set device to cpu!'
            for k, m in self.modules.items():
                self.modules[k] = m.to(torch.device('cuda', device_ids[0]))
            self.device = torch.device('cuda', device_ids[0])

        # ddp trainer
        if len(device_ids) > 1:
            assert torch.cuda.is_available(
            ), 'GPU is not available, please set device to cpu!'
            assert torch.cuda.device_count() > max(
                device_ids
            ), f'{torch.cuda.device_count()} GPUs are available, but try to access f{max(device_ids)}-th GPU!'

            rank = int(os.getenv('RANK', -1))
            local_rank = int(os.getenv('LOCAL_RANK', -1))
            world_size = int(os.getenv('WORLD_SIZE', 1))

            assert rank != -1, 'TorchEEG support to use multiple GPU with distributed data parallel, and you need to run the python script with torch.distributed.launch or torchrun.'

            self.rank = rank
            self.local_rank = local_rank
            self.world_size = world_size

            os.environ["CUDA_VISIBLE_DEVICES"] = ",".join(
                [str(d) for d in device_ids])

            assert dist.is_available(
            ), 'Distributed data parallel is not supported!'
            if not dist.is_initialized():
                # nccl is not available in windows
                backend = "nccl" if dist.is_nccl_available() else "gloo"
                if self.rank == 0:
                    self.log(f"Using backend: {backend}.")
                dist.init_process_group(backend=backend,
                                        rank=self.rank,
                                        world_size=self.world_size)

            for k, m in self.modules.items():
                self.modules[k] = m.to(torch.device('cuda', local_rank))

            for k, m in self.modules.items():
                ddp_m = DistributedDataParallel(m,
                                                device_ids=[self.local_rank],
                                                find_unused_parameters=True)
                if self.ddp_sync_bn:
                    ddp_m = nn.SyncBatchNorm.convert_sync_batchnorm(ddp_m)
                self.modules[k] = ddp_m

            self.device = torch.device(self.local_rank)

    @property
    def is_distributed(self):
        return self.rank != -1

    @property
    def is_main(self):
        # if not distributed mode -1, then it is the main process
        # if distributed mode > -1, then 0 is the main process
        return self.rank in {-1, 0}

    def log(self, *args, **kwargs):
        # can be overwritten
        if self.is_main:
            print(*args, **kwargs)

    def before_training_epoch(self, epoch_id: int, num_epochs: int):
        # can be overwritten
        ...

    def before_training_step(self, batch_id: int, num_batches: int):
        # can be overwritten
        ...

    def on_training_step(self, train_batch: Tuple, batch_id: int,
                         num_batches: int):
        raise NotImplementedError

    def after_training_step(self, batch_id: int, num_batches: int):
        # can be overwritten
        ...

    def after_training_epoch(self, epoch_id: int, num_epochs: int):
        # can be overwritten
        ...

    def before_validation_epoch(self, epoch_id: int, num_epochs: int):
        # can be overwritten
        ...

    def before_validation_step(self, batch_id: int, num_batches: int):
        # can be overwritten
        ...

    def on_validation_step(self, val_batch: Tuple, batch_id: int,
                           num_batches: int):
        raise NotImplementedError

    def after_validation_step(self, batch_id: int, num_batches: int):
        # can be overwritten
        ...

    def after_validation_epoch(self, epoch_id: int, num_epochs: int):
        # can be overwritten
        ...

    def on_reveive_dataloader(self, dataloader, mode='train'):
        if mode == 'test' and not self.ddp_test:
            return dataloader
        if mode == 'val' and not self.ddp_val:
            return dataloader

        if not self.is_distributed:
            return dataloader

        shuffle = True
        if isinstance(dataloader.sampler, SequentialSampler):
            shuffle = False
        sampler = DistributedSampler(dataloader.dataset, shuffle=shuffle)
        dataloader = DataLoader(dataloader.dataset,
                                dataloader.batch_size,
                                sampler=sampler,
                                num_workers=dataloader.num_workers,
                                pin_memory=dataloader.pin_memory,
                                drop_last=dataloader.drop_last,
                                collate_fn=dataloader.collate_fn)

        dataloader.need_to_set_epoch = True
        return dataloader

    def fit(self,
            train_loader: DataLoader,
            val_loader: DataLoader,
            num_epochs: int = 1,
            **kwargs):
        r'''
        Args:
            train_loader (DataLoader): Iterable DataLoader for traversing the training data batch (torch.utils.data.dataloader.DataLoader, torch_geometric.loader.DataLoader, etc).
            val_loader (DataLoader): Iterable DataLoader for traversing the validation data batch (torch.utils.data.dataloader.DataLoader, torch_geometric.loader.DataLoader, etc).
            num_epochs (int): training epochs. (defualt: :obj:`1`)
        '''
        train_loader = self.on_reveive_dataloader(train_loader, mode='train')
        val_loader = self.on_reveive_dataloader(val_loader, mode='val')

        for t in range(num_epochs):
            if hasattr(train_loader, 'need_to_set_epoch'):
                train_loader.sampler.set_epoch(t)
            if hasattr(val_loader, 'need_to_set_epoch'):
                val_loader.sampler.set_epoch(t)

            num_batches = len(train_loader)

            # set model to train mode
            for k, m in self.modules.items():
                self.modules[k].train()

            # hook
            self.before_training_epoch(t + 1, num_epochs, **kwargs)
            for batch_id, train_batch in enumerate(train_loader):
                # hook
                self.before_training_step(batch_id, num_batches, **kwargs)
                # hook
                self.on_training_step(train_batch, batch_id, num_batches,
                                      **kwargs)
                # hook
                self.after_training_step(batch_id, num_batches, **kwargs)
            # hook
            self.after_training_epoch(t + 1, num_epochs, **kwargs)

            # set model to val mode
            for k, m in self.modules.items():
                self.modules[k].eval()

            num_batches = len(val_loader)
            # hook
            self.before_validation_epoch(t + 1, num_epochs, **kwargs)
            with torch.no_grad():
                for batch_id, val_batch in enumerate(val_loader):
                    # hook
                    self.before_validation_step(batch_id, num_batches, **kwargs)
                    # hook
                    self.on_validation_step(val_batch, batch_id, num_batches,
                                            **kwargs)
                    # hook
                    self.after_validation_step(batch_id, num_batches, **kwargs)
                    # hook
            self.after_validation_epoch(t + 1, num_epochs, **kwargs)

        return self

    def test(self, test_loader: DataLoader, **kwargs):
        r'''
        Args:
            test_loader (DataLoader): Iterable DataLoader for traversing the test data batch (torch.utils.data.dataloader.DataLoader, torch_geometric.loader.DataLoader, etc).
        '''
        test_loader = self.on_reveive_dataloader(test_loader, mode='test')

        for k, m in self.modules.items():
            self.modules[k].eval()

        num_batches = len(test_loader)
        self.before_test_epoch(**kwargs)
        with torch.no_grad():
            for batch_id, test_batch in enumerate(test_loader):
                # hook
                self.before_test_step(batch_id, num_batches, **kwargs)
                # hook
                self.on_test_step(test_batch, batch_id, num_batches, **kwargs)
                # hook
                self.after_test_step(batch_id, num_batches, **kwargs)
        self.after_test_epoch(**kwargs)

    def before_test_epoch(self, **kwargs):
        # can be overwritten
        ...

    def before_test_step(self, batch_id: int, num_batches: int, **kwargs):
        # can be overwritten
        ...

    def on_test_step(self, test_batch: Tuple, batch_id: int, num_batches: int, **kwargs):
        raise NotImplementedError

    def after_test_step(self, batch_id: int, num_batches: int, **kwargs):
        # can be overwritten
        ...

    def after_test_epoch(self, **kwargs):
        # can be overwritten
        ...

    def load_state_dict(self, load_path, strict=False):
        if self.is_distributed:
            map_location = {'cpu': f'cuda:{self.local_rank}'}
        else:
            map_location = {'cpu': self.device}

        state_dict = torch.load(load_path, map_location=map_location)
        for k, m in self.modules.items():
            self.modules[k].load_state_dict(state_dict[k], strict=strict)

    def save_state_dict(self, save_path):
        if not os.path.exists(os.path.dirname(save_path)):
            os.makedirs(os.path.dirname(save_path))

        if self.is_main:
            state_dict = {}
            for k, m in self.modules.items():
                if isinstance(m, DistributedDataParallel):
                    m = m.module
                state_dict[k] = m.state_dict()
            torch.save(state_dict, save_path)

        # Use a barrier() to make sure that process a loads the model after process b
        # saves it (In case the user calls load_state_dict right after save_state_dict).
        if self.is_distributed:
            dist.barrier()


import math
import os
import shutil
import pandas as pd
from functools import partial
from multiprocessing import Manager, Pool, Process, Queue
from typing import Callable, Dict, List, Union

from torch.utils.data import Dataset
from torcheeg.io import EEGSignalIO, MetaInfoIO
from tqdm import tqdm

MAX_QUEUE_SIZE = 1024


def copyfile(src, dst, *, follow_symlinks=True):
    if shutil._samefile(src, dst):
        raise shutil.SameFileError("{!r} and {!r} are the same file".format(
            src, dst))

    for fn in [src, dst]:
        try:
            st = os.stat(fn)
        except OSError:
            pass
        else:
            if shutil.stat.S_ISFIFO(st.st_mode):
                raise shutil.SpecialFileError("`%s` is a named pipe" % fn)

    if not follow_symlinks and os.path.islink(src):
        os.symlink(os.readlink(src), dst)
    else:
        size = os.stat(src).st_size
        with open(src, 'rb') as fsrc:
            with open(dst, 'wb') as fdst:
                copyfileobj(fsrc, fdst, total=size)
    return dst


def copyfileobj(fsrc, fdst, total, length=16 * 1024):
    pbar = tqdm(total=total)
    pbar.set_description("[FROM EXISTING]")
    while True:
        buf = fsrc.read(length)
        if not buf:
            break
        fdst.write(buf)
        pbar.update(len(buf))


def copy_with_progress(src, dst, *, follow_symlinks=True):
    if os.path.isdir(dst):
        dst = os.path.join(dst, os.path.basename(src))
    copyfile(src, dst, follow_symlinks=follow_symlinks)
    shutil.copymode(src, dst)
    return dst


def split_df_by_size(df, chunk_size=100):
    chunks = list()
    num_chunks = math.ceil(len(df) / chunk_size)
    for i in range(num_chunks):
        chunks.append(df[i * chunk_size:(i + 1) * chunk_size])
    return chunks


def split_df_by_column(df, column_name='baseline_id'):
    chunks = list()
    column_uniques = df[column_name].unique()
    for column in column_uniques:
        chunks.append(df[df[column_name].isin([column])])
    return chunks


def io_consumer(write_eeg_fn, queue):
    while True:
        item = queue.get()
        if not item is None:
            eeg = item['eeg']
            key = item['key']
            write_eeg_fn(eeg, key)
        else:
            break


class SingleProcessingQueue:
    def __init__(self, write_eeg_fn):
        self.write_eeg_fn = write_eeg_fn

    def put(self, item):
        eeg = item['eeg']
        key = item['key']
        self.write_eeg_fn(eeg, key)


def basic_producer(chunk: pd.DataFrame, transform: Callable,
                   read_eeg_fn: Callable, queue: Queue):

    last_baseline_id = None
    last_baseline_sample = None

    for i, row in chunk.iterrows():
        clip_id = row['clip_id']
        clip_sample = read_eeg_fn(clip_id)

        if 'baseline_id' in row:
            baseline_id = row['baseline_id']

            if last_baseline_id == baseline_id:
                trial_baseline_sample = last_baseline_sample
            else:
                trial_baseline_sample = read_eeg_fn(baseline_id)

            t = transform(eeg=clip_sample, baseline=trial_baseline_sample)
            t_eeg = t['eeg']
            t_baseline = t['baseline']

            if not last_baseline_id == baseline_id:
                queue.put({'eeg': t_baseline, 'key': baseline_id})

                last_baseline_id = baseline_id
                last_baseline_sample = trial_baseline_sample

            queue.put({'eeg': t_eeg, 'key': clip_id})

        else:
            t = transform(eeg=clip_sample)
            t_eeg = t['eeg']
            queue.put({'eeg': t_eeg, 'key': clip_id})


def reduce_producer(chunk: List[pd.DataFrame], transform: Callable,
                    read_eeg_fn: Callable, write_info_fn: Callable,
                    queue: Queue):

    for reduce_list in chunk:

        reduce_info = None

        reduce_clip_list = []
        reduce_baseline_list = []

        reduce_list = reduce_list.sort_values("clip_id", inplace=False)

        for i, row in reduce_list.iterrows():
            if reduce_info is None:
                reduce_info = row.to_dict()
            reduce_clip_list.append(read_eeg_fn(row['clip_id']))
            if 'baseline_id' in row:
                reduce_baseline_list.append(read_eeg_fn(row['baseline_id']))

        reduce_clip = transform(reduce_clip_list)
        queue.put({'eeg': reduce_clip, 'key': reduce_info['clip_id']})

        if len(reduce_baseline_list):
            reduce_baseline = transform(reduce_baseline_list)
            queue.put({
                'eeg': reduce_baseline,
                'key': reduce_info['baseline_id']
            })

        write_info_fn(reduce_info)


class BaseDataset(Dataset):
    channel_location_dict = {}
    adjacency_matrix = []

    def __init__(self, io_path: str):
        if not self.exist(io_path):
            raise RuntimeError(
                'Database IO does not exist, please regenerate database IO.')
        self.io_path = io_path

        meta_info_io_path = os.path.join(self.io_path, 'info.csv')
        eeg_signal_io_path = os.path.join(self.io_path, 'eeg')

        info_io = MetaInfoIO(meta_info_io_path)
        self.eeg_io = EEGSignalIO(eeg_signal_io_path)

        self.info = info_io.read_all()

    def exist(self, io_path: str) -> bool:
        meta_info_io_path = os.path.join(io_path, 'info.csv')
        eeg_signal_io_path = eeg_signal_io_path = os.path.join(io_path, 'eeg')

        return os.path.exists(meta_info_io_path) and os.path.exists(
            eeg_signal_io_path)

    def __getitem__(self, index: int) -> any:
        raise NotImplementedError(
            "Method __getitem__ is not implemented in class " +
            self.__class__.__name__)

    def __len__(self):
        return len(self.info)

    def __copy__(self) -> 'BaseDataset':
        cls = self.__class__
        result = cls.__new__(cls)
        result.__dict__.update(self.__dict__)
        eeg_signal_io_path = os.path.join(self.io_path, 'eeg')

        result.eeg_io = EEGSignalIO(eeg_signal_io_path)
        return result

    @property
    def repr_body(self) -> Dict:
        return {'io_path': self.io_path}

    @property
    def repr_tail(self) -> Dict:
        return {'length': self.__len__()}

    def __repr__(self) -> str:
        # init info
        format_string = self.__class__.__name__ + '('
        for i, (k, v) in enumerate(self.repr_body.items()):
            # line end
            if i:
                format_string += ','
            format_string += '\n'
            # str param
            if isinstance(v, str):
                format_string += f"    {k}='{v}'"
            else:
                format_string += f"    {k}={v}"
        format_string += '\n)'
        # other info
        format_string += '\n'
        for i, (k, v) in enumerate(self.repr_tail.items()):
            if i:
                format_string += ', '
            format_string += f'{k}={v}'
        return format_string

    @staticmethod
    def from_existing(dataset,
                      io_path: str,
                      offline_transform: Union[None, Callable] = None,
                      num_worker: int = 0,
                      verbose: bool = True,
                      cache_size: int = 64 * 1024 * 1024 * 1024,
                      chunk_size_for_worker: int = 100,
                      chunk_column_for_worker: str = 'baseline_id',
                      **args):
        meta_info_io_path = os.path.join(io_path, 'info.csv')
        eeg_signal_io_path = eeg_signal_io_path = os.path.join(io_path, 'eeg')

        exist_io = os.path.exists(meta_info_io_path) and os.path.exists(
            eeg_signal_io_path)

        if not exist_io:
            os.makedirs(io_path, exist_ok=True)

            # copy info file
            shutil.copy(os.path.join(dataset.io_path, 'info.csv'),
                        os.path.join(io_path, 'info.csv'))
            # copy dataset
            if not offline_transform is None:
                eeg_io = EEGSignalIO(eeg_signal_io_path, cache_size=cache_size)
                if chunk_column_for_worker in dataset.info.columns:
                    chunk_for_worker = split_df_by_column(
                        dataset.info, chunk_column_for_worker)
                else:
                    chunk_for_worker = split_df_by_size(dataset.info,
                                                        chunk_size_for_worker)

                if verbose:
                    # show process bar
                    pbar = tqdm(total=len(chunk_for_worker))
                    pbar.set_description("[FROM EXISTING]")
                if num_worker > 1:
                    manager = Manager()
                    queue = manager.Queue(maxsize=MAX_QUEUE_SIZE)

                    partial_mp_fn = partial(basic_producer,
                                            transform=offline_transform,
                                            read_eeg_fn=dataset.eeg_io.read_eeg,
                                            queue=queue)

                    io_consumer_process = Process(target=io_consumer,
                                                  args=(eeg_io.write_eeg,
                                                        queue),
                                                  daemon=True)
                    io_consumer_process.start()

                    for _ in Pool(num_worker).imap(partial_mp_fn,
                                                   chunk_for_worker):
                        if verbose:
                            pbar.update(1)

                    queue.put(None)

                    io_consumer_process.join()
                    io_consumer_process.close()

                else:
                    for chunk in chunk_for_worker:
                        basic_producer(chunk=chunk,
                                       transform=offline_transform,
                                       read_eeg_fn=dataset.eeg_io.read_eeg,
                                       queue=SingleProcessingQueue(
                                           eeg_io.write_eeg))
                        if verbose:
                            pbar.update(1)

                if verbose:
                    pbar.close()
                    print('Please wait for the writing process to complete...')
            else:
                if verbose:
                    os.makedirs(os.path.join(io_path, 'eeg'), exist_ok=True)
                    copy_with_progress(
                        os.path.join(dataset.io_path, 'eeg', 'data.mdb'),
                        os.path.join(io_path, 'eeg', 'data.mdb'))
                else:
                    shutil.copytree(os.path.join(dataset.io_path, 'eeg'),
                                    os.path.join(io_path, 'eeg'))

        return type(dataset)(io_path=io_path,
                             offline_transform=offline_transform,
                             num_worker=num_worker,
                             verbose=verbose,
                             cache_size=cache_size,
                             **args)

    @staticmethod
    def reduce_from_existing(dataset,
                             io_path: str,
                             reduce_fn: Callable,
                             reduce_by: str = 'epoch_id',
                             verbose: bool = True,
                             cache_size: int = 64 * 1024 * 1024 * 1024,
                             num_worker: int = 0,
                             chunk_size_for_worker: int = 100,
                             **args):
        meta_info_io_path = os.path.join(io_path, 'info.csv')
        eeg_signal_io_path = eeg_signal_io_path = os.path.join(io_path, 'eeg')

        exist_io = os.path.exists(meta_info_io_path) and os.path.exists(
            eeg_signal_io_path)

        if not exist_io:
            os.makedirs(io_path, exist_ok=True)

            info_io = MetaInfoIO(meta_info_io_path)
            eeg_io = EEGSignalIO(eeg_signal_io_path, cache_size=cache_size)

            reduce_list = split_df_by_column(dataset.info, reduce_by)
            chunk_for_worker = [
                reduce_list[i:i + chunk_size_for_worker]
                for i in range(0, len(reduce_list), chunk_size_for_worker)
            ]

            if verbose:
                # show process bar
                pbar = tqdm(total=len(chunk_for_worker))
                pbar.set_description("[REDUCE]")

            if num_worker > 1:
                manager = Manager()
                queue = manager.Queue(maxsize=MAX_QUEUE_SIZE)
                io_consumer_process = Process(target=io_consumer,
                                              args=(eeg_io.write_eeg, queue),
                                              daemon=True)
                io_consumer_process.start()

                partial_mp_fn = partial(reduce_producer,
                                        transform=reduce_fn,
                                        read_eeg_fn=dataset.eeg_io.read_eeg,
                                        write_info_fn=info_io.write_info,
                                        queue=queue)

                for _ in Pool(num_worker).imap(partial_mp_fn, chunk_for_worker):
                    if verbose:
                        pbar.update(1)

                queue.put(None)

                io_consumer_process.join()
                io_consumer_process.close()
            else:
                for chunk in chunk_for_worker:
                    reduce_producer(chunk=chunk,
                                    transform=reduce_fn,
                                    read_eeg_fn=dataset.eeg_io.read_eeg,
                                    write_info_fn=info_io.write_info,
                                    queue=SingleProcessingQueue(
                                        eeg_io.write_eeg))
                    if verbose:
                        pbar.update(1)

            if verbose:
                pbar.close()

        return type(dataset)(io_path=io_path,
                             verbose=verbose,
                             cache_size=cache_size,
                             num_worker=num_worker,
                             **args)



import glob
import json
import os
from functools import partial
from multiprocessing import Manager, Pool, Process, Queue, set_start_method
from typing import Callable, List, Union

import mne
import xmltodict
from torcheeg.io import EEGSignalIO, MetaInfoIO
from tqdm import tqdm

MAX_QUEUE_SIZE = 1024


def transform_producer(file_name: str, root_path: str, chunk_size: int,
                       sampling_rate: int, overlap: int, num_channel: int,
                       num_baseline: int, baseline_chunk_size: int,
                       num_trial_sample: int, before_trial: Union[Callable,
                                                                  None],
                       transform: Union[Callable,
                                        None], after_trial: Union[Callable,
                                                                  None],
                       write_info_fn: Callable, queue: Queue):
    trial_dir = os.path.join(root_path, file_name)

    mne.set_log_level('CRITICAL')
    # record the common meta info for the trial
    label_file = os.path.join(trial_dir, 'session.xml')
    emodims = [
        '@feltArsl', '@feltCtrl', '@feltEmo', '@feltPred', '@feltVlnc',
        '@isStim'
    ]
    with open(label_file) as f:
        label_info = xmltodict.parse('\n'.join(f.readlines()))
    label_info = json.loads(json.dumps(label_info))['session']

    if not '@feltArsl' in label_info:
        # skip label_info['@isStim'] == '0' and other exception
        return

    trial_meta_info = {
        'subject_id': label_info['subject']['@id'],
        'trial_id': label_info['@mediaFile'],
        'duration': float(label_info['@cutLenSec'])
    }
    # feltArsl, feltCtrl, feltEmo, feltPred, feltVlnc, isStim
    trial_meta_info.update({k[1:]: int(label_info[k]) for k in emodims})

    write_pointer = 0

    # extract signals
    sample_file = glob.glob(str(os.path.join(trial_dir, '*.bdf')))[0]

    raw = mne.io.read_raw_bdf(sample_file, preload=True, stim_channel='Status')
    events = mne.find_events(raw, stim_channel='Status')

    montage = mne.channels.make_standard_montage(kind='biosemi32')
    raw.set_montage(montage, on_missing='ignore')

    # pick channels
    raw.pick_channels(raw.ch_names[:num_channel])

    start_samp, end_samp = events[0][0] + 1, events[1][0] - 1

    # extract baseline signals
    trial_baseline_raw = raw.copy().crop(raw.times[0], raw.times[end_samp])
    trial_baseline_raw = trial_baseline_raw.resample(sampling_rate)

    trial_baseline_sample = trial_baseline_raw.to_data_frame().to_numpy(
    )[:, 1:].swapaxes(1, 0)  # channel(32), timestep(30 * 128)
    trial_baseline_sample = trial_baseline_sample[:, :num_baseline *
                                                  baseline_chunk_size]
    trial_baseline_sample = trial_baseline_sample.reshape(
        num_channel, num_baseline,
        baseline_chunk_size).mean(axis=1)  # channel(32), timestep(128)

    # extract experimental signals
    trial_raw = raw.copy().crop(raw.times[start_samp], raw.times[end_samp])
    trial_raw = trial_raw.resample(sampling_rate)
    trial_samples = trial_raw.to_data_frame().to_numpy()[:, 1:].swapaxes(1, 0)
    if before_trial:
        trial_samples = before_trial(trial_samples)

    start_at = 0
    if chunk_size <= 0:
        chunk_size = trial_samples.shape[1] - start_at

    # chunk with chunk size
    end_at = chunk_size
    # calculate moving step
    step = chunk_size - overlap

    max_len = trial_samples.shape[1]
    if not (num_trial_sample <= 0):
        max_len = min(num_trial_sample * chunk_size, trial_samples.shape[1])

    trial_queue = []
    while end_at <= max_len:
        clip_sample = trial_samples[:, start_at:end_at]

        t_eeg = clip_sample
        t_baseline = trial_baseline_sample
        if not transform is None:
            t = transform(eeg=clip_sample, baseline=trial_baseline_sample)
            t_eeg = t['eeg']
            t_baseline = t['baseline']

        # put baseline signal into IO
        if not 'baseline_id' in trial_meta_info:
            trial_base_id = f'{file_name}_{write_pointer}'

            queue.put({'eeg': t_baseline, 'key': trial_base_id})
            write_pointer += 1
            trial_meta_info['baseline_id'] = trial_base_id

        clip_id = f'{file_name}_{write_pointer}'
        if after_trial:
            trial_queue.append({'eeg': t_eeg, 'key': clip_id})
        else:
            queue.put({'eeg': t_eeg, 'key': clip_id})
        write_pointer += 1

        # record meta info for each signal
        record_info = {
            'start_at': start_at,
            'end_at': end_at,
            'clip_id': clip_id
        }
        record_info.update(trial_meta_info)
        write_info_fn(record_info)
        start_at = start_at + step
        end_at = start_at + chunk_size

    if len(trial_queue) and after_trial:
        trial_queue = after_trial(trial_queue)
        for obj in trial_queue:
            assert 'eeg' in obj and 'key' in obj, 'after_trial must return a list of dictionaries, where each dictionary corresponds to an EEG sample, containing `eeg` and `key` as keys.'
            queue.put(obj)


def io_consumer(write_eeg_fn, queue):
    while True:
        item = queue.get()
        if not item is None:
            eeg = item['eeg']
            key = item['key']
            write_eeg_fn(eeg, key)
        else:
            break


class SingleProcessingQueue:
    def __init__(self, write_eeg_fn):
        self.write_eeg_fn = write_eeg_fn

    def put(self, item):
        eeg = item['eeg']
        key = item['key']
        self.write_eeg_fn(eeg, key)


def mahnob_constructor(root_path: str = './Sessions',
                       chunk_size: int = 128,
                       sampling_rate: int = 128,
                       overlap: int = 0,
                       num_channel: int = 32,
                       num_baseline: int = 30,
                       baseline_chunk_size: int = 128,
                       num_trial_sample: int = 30,
                       before_trial: Union[None, Callable] = None,
                       transform: Union[None, Callable] = None,
                       after_trial: Union[Callable, None] = None,
                       io_path: str = './io/mahnob',
                       num_worker: int = 0,
                       verbose: bool = True,
                       cache_size: int = 64 * 1024 * 1024 * 1024) -> None:
    # init IO
    meta_info_io_path = os.path.join(io_path, 'info.csv')
    eeg_signal_io_path = os.path.join(io_path, 'eeg')

    if os.path.exists(io_path) and not os.path.getsize(meta_info_io_path) == 0:
        print(
            f'The target folder already exists, if you need to regenerate the database IO, please delete the path {io_path}.'
        )
        return

    os.makedirs(io_path, exist_ok=True)

    info_io = MetaInfoIO(meta_info_io_path)
    eeg_io = EEGSignalIO(eeg_signal_io_path, cache_size=cache_size)

    # loop to access the dataset files
    file_list = os.listdir(root_path)

    if verbose:
        # show process bar
        pbar = tqdm(total=len(file_list))
        pbar.set_description("[MAHNOB]")

    if num_worker > 1:
        manager = Manager()
        queue = manager.Queue(maxsize=MAX_QUEUE_SIZE)
        io_consumer_process = Process(target=io_consumer,
                                      args=(eeg_io.write_eeg, queue),
                                      daemon=True)
        io_consumer_process.start()

        partial_mp_fn = partial(transform_producer,
                                root_path=root_path,
                                chunk_size=chunk_size,
                                sampling_rate=sampling_rate,
                                overlap=overlap,
                                num_channel=num_channel,
                                num_baseline=num_baseline,
                                baseline_chunk_size=baseline_chunk_size,
                                num_trial_sample=num_trial_sample,
                                before_trial=before_trial,
                                transform=transform,
                                after_trial=after_trial,
                                write_info_fn=info_io.write_info,
                                queue=queue)

        for _ in Pool(num_worker).imap(partial_mp_fn, file_list):
            if verbose:
                pbar.update(1)

        queue.put(None)

        io_consumer_process.join()
        io_consumer_process.close()

    else:
        for file_name in file_list:
            transform_producer(file_name=file_name,
                               root_path=root_path,
                               chunk_size=chunk_size,
                               sampling_rate=sampling_rate,
                               overlap=overlap,
                               num_channel=num_channel,
                               num_baseline=num_baseline,
                               baseline_chunk_size=baseline_chunk_size,
                               num_trial_sample=num_trial_sample,
                               before_trial=before_trial,
                               transform=transform,
                               after_trial=after_trial,
                               write_info_fn=info_io.write_info,
                               queue=SingleProcessingQueue(eeg_io.write_eeg))
            if verbose:
                pbar.update(1)

    if verbose:
        pbar.close()
        print('Please wait for the writing process to complete...')



import pickle
import os
from typing import Union

import torch
import lmdb


class EEGSignalIO:
    r'''
    A general-purpose, lightweight and efficient EEG signal IO APIs for converting various real-world EEG signal datasets into samples and storing them in the database. Here, we draw on the implementation ideas of industrial-grade application Caffe, and encapsulate a set of EEG signal reading and writing methods based on Lightning Memory-Mapped Database (LMDB), which not only unifies the differences of data types in different databases, but also accelerates the reading of data during training and testing.

    .. code-block:: python

        eeg_io = EEGSignalIO('YOUR_PATH')
        key = eeg_io.write_eeg(np.random.randn(32, 128))
        eeg = eeg_io.read_eeg(key)
        eeg.shape
        >>> (32, 128)
    
    Args:
        cache_path (str): Where the database is stored.
        cache_size (int): The maximum capacity of the database. (default: :obj:`1099511627776`)
    '''
    def __init__(self, cache_path: str, cache_size: int = 1024 * 1024 * 1024 * 1024) -> None:
        self.cache_path = cache_path
        self.cache_size = cache_size

        if not os.path.exists(self.cache_path):
            os.makedirs(self.cache_path, exist_ok=True)

    @property
    def write_pointer(self):
        return len(self)

    def __len__(self):
        with lmdb.open(self.cache_path, self.cache_size, lock=False) as env:
            with env.begin() as transaction:
                return transaction.stat()['entries']

    def write_eeg(self, eeg: Union[any, torch.Tensor], key: Union[str, None] = None) -> str:
        r'''
        Write EEG signal to database.

        Args:
            eeg (any): EEG signal samples to be written into the database.
            key (str, optional): The key of the EEG signal to be inserted, if not specified, it will be an auto-incrementing integer.

        Returns:
            int: The index of written EEG signals in the database.
        '''

        if key is None:
            key = str(self.write_pointer)

        if eeg is None:
            raise RuntimeError(f'Save None to the LMDB with the key {key}!')

        with lmdb.open(self.cache_path, self.cache_size, lock=False) as env:
            with env.begin(write=True) as transaction:
                transaction.put(key.encode(), pickle.dumps(eeg))
            return key

    def read_eeg(self, key: str) -> any:
        r'''
        Query the corresponding EEG signal in the database according to the index.

        Args:
            key (str): The index of the EEG signal to be queried.
            
        Returns:
            any: The EEG signal sample.
        '''

        with lmdb.open(self.cache_path, self.cache_size, lock=False) as env:
            with env.begin() as transaction:
                eeg = transaction.get(key.encode())

            if eeg is None:
                raise RuntimeError(f'Unable to index the EEG signal sample with key {key}!')

            return pickle.loads(eeg)


import os
import re
import numpy as np

from functools import partial
from multiprocessing import Manager, Pool, Process, Queue
from typing import Callable, Union

import scipy.io as scio
from torcheeg.io import EEGSignalIO, MetaInfoIO
from tqdm import tqdm

MAX_QUEUE_SIZE = 1024


def transform_producer(file_path: str, feature: list, num_channel: int,
                       before_trial: Union[None, Callable],
                       transform: Union[None,
                                        Callable], after_trial: Union[Callable,
                                                                      None],
                       write_info_fn: Callable, queue: Queue):
    session_id = os.path.basename(os.path.dirname(file_path))
    _, file_name = os.path.split(file_path)

    subject = int(
        os.path.basename(file_name).split('.')[0].split('_')[0])  # subject (15)
    date = int(
        os.path.basename(file_name).split('.')[0].split('_')[1])  # period (3)

    samples = scio.loadmat(file_path, verify_compressed_data_integrity=False
                           )  # trial (15), channel(62), timestep(n*200)
    # label file
    labels = [[
        1, 2, 3, 0, 2, 0, 0, 1, 0, 1, 2, 1, 1, 1, 2, 3, 2, 2, 3, 3, 0, 3, 0, 3
    ], [
        2, 1, 3, 0, 0, 2, 0, 2, 3, 3, 2, 3, 2, 0, 1, 1, 2, 1, 0, 3, 0, 1, 3, 1
    ], [
        1, 2, 2, 1, 3, 3, 3, 1, 1, 2, 1, 0, 2, 3, 3, 0, 2, 3, 0, 0, 2, 0, 1, 0
    ]]  # The labels with 0, 1, 2, and 3 denote the ground truth, neutral, sad, fear, and happy emotions, respectively.
    session_labels = labels[int(session_id) - 1]

    trial_ids = [
        int(re.findall(r"de_movingAve(\d+)", key)[0]) for key in samples.keys()
        if 'de_movingAve' in key
    ]

    write_pointer = 0
    # loop for each trial
    for trial_id in trial_ids:
        # extract baseline signals
        trial_samples = []
        for cur_feature in feature:
            trial_samples.append(
                samples[cur_feature +
                        str(trial_id)])  # channel(61), timestep(n), bands(5)
        trial_samples = np.concatenate(
            trial_samples,
            axis=-1)[:num_channel]  # channel(61), timestep(n), features(5*k)
        trial_samples = trial_samples.transpose((1, 0, 2))
        # timestep(n), channel(61), features(5*k)

        if before_trial:
            trial_samples = before_trial(trial_samples)

        # record the common meta info
        trial_meta_info = {
            'subject_id': subject,
            'trial_id': trial_id,
            'session_id': session_id,
            'emotion': int(session_labels[trial_id - 1]),
            'date': date
        }

        trial_queue = []
        for i, clip_sample in enumerate(trial_samples):
            t_eeg = clip_sample
            if not transform is None:
                t_eeg = transform(eeg=clip_sample)['eeg']

            clip_id = f'{file_name}_{write_pointer}'
            if after_trial:
                trial_queue.append({'eeg': t_eeg, 'key': clip_id})
            else:
                queue.put({'eeg': t_eeg, 'key': clip_id})
            write_pointer += 1

            # record meta info for each signal
            record_info = {
                'start_at': i * 400,
                'end_at': (i + 1) *
                400,  # The size of the sliding time windows for feature extraction is 4 seconds.
                'clip_id': clip_id
            }
            record_info.update(trial_meta_info)
            write_info_fn(record_info)

        if len(trial_queue) and after_trial:
            trial_queue = after_trial(trial_queue)
            for obj in trial_queue:
                assert 'eeg' in obj and 'key' in obj, 'after_trial must return a list of dictionaries, where each dictionary corresponds to an EEG sample, containing `eeg` and `key` as keys.'
                queue.put(obj)


def io_consumer(write_eeg_fn, queue):
    while True:
        item = queue.get()
        if not item is None:
            eeg = item['eeg']
            key = item['key']
            write_eeg_fn(eeg, key)
        else:
            break


class SingleProcessingQueue:
    def __init__(self, write_eeg_fn):
        self.write_eeg_fn = write_eeg_fn

    def put(self, item):
        eeg = item['eeg']
        key = item['key']
        self.write_eeg_fn(eeg, key)


def seed_iv_feature_constructor(
        root_path: str = './eeg_feature_smooth',
        feature: list = ['de_movingAve1'],
        num_channel: int = 61,
        before_trial: Union[None, Callable] = None,
        transform: Union[None, Callable] = None,
        after_trial: Union[Callable, None] = None,
        io_path: str = './io/seed_iv_feature',
        num_worker: int = 0,
        verbose: bool = True,
        cache_size: int = 64 * 1024 * 1024 * 1024) -> None:
    # init IO
    meta_info_io_path = os.path.join(io_path, 'info.csv')
    eeg_signal_io_path = os.path.join(io_path, 'eeg')

    if os.path.exists(io_path) and not os.path.getsize(meta_info_io_path) == 0:
        print(
            f'The target folder already exists, if you need to regenerate the database IO, please delete the path {io_path}.'
        )
        return

    os.makedirs(io_path, exist_ok=True)

    meta_info_io_path = os.path.join(io_path, 'info.csv')
    eeg_signal_io_path = os.path.join(io_path, 'eeg')

    info_io = MetaInfoIO(meta_info_io_path)
    eeg_io = EEGSignalIO(eeg_signal_io_path, cache_size=cache_size)

    # loop to access the dataset files
    session_list = ['1', '2', '3']
    file_path_list = []
    for session in session_list:
        session_root_path = os.path.join(root_path, session)
        for file_name in os.listdir(session_root_path):
            file_path_list.append(os.path.join(session_root_path, file_name))

    if verbose:
        # show process bar
        pbar = tqdm(total=len(file_path_list))
        pbar.set_description("[SEED-IV FEATURE]")

    if num_worker > 1:
        manager = Manager()
        queue = manager.Queue(maxsize=MAX_QUEUE_SIZE)
        io_consumer_process = Process(target=io_consumer,
                                      args=(eeg_io.write_eeg, queue),
                                      daemon=True)
        io_consumer_process.start()

        partial_mp_fn = partial(transform_producer,
                                feature=feature,
                                num_channel=num_channel,
                                before_trial=before_trial,
                                transform=transform,
                                after_trial=after_trial,
                                write_info_fn=info_io.write_info,
                                queue=queue)

        for _ in Pool(num_worker).imap(partial_mp_fn, file_path_list):
            if verbose:
                pbar.update(1)

        queue.put(None)

        io_consumer_process.join()
        io_consumer_process.close()

    else:
        for file_path in file_path_list:
            transform_producer(file_path=file_path,
                               feature=feature,
                               num_channel=num_channel,
                               before_trial=before_trial,
                               transform=transform,
                               after_trial=after_trial,
                               write_info_fn=info_io.write_info,
                               queue=SingleProcessingQueue(eeg_io.write_eeg))
            if verbose:
                pbar.update(1)

    if verbose:
        pbar.close()
        print('Please wait for the writing process to complete...')



import os
from functools import partial
from multiprocessing import Manager, Pool, Process, Queue
from typing import Callable, List, Union

import joblib
from torcheeg.io import EEGSignalIO, MetaInfoIO
from tqdm import tqdm

MAX_QUEUE_SIZE = 1024

FIRST_BATCH_CHANNEL = [
    'FP1', 'FP2', 'FZ', 'F3', 'F4', 'F7', 'F8', 'FC1', 'FC2', 'FC5', 'FC6',
    'CZ', 'C3', 'C4', 'T7', 'T8', 'A1', 'A2', 'CP1', 'CP2', 'CP5', 'CP6', 'PZ',
    'P3', 'P4', 'P7', 'P8', 'PO3', 'PO4', 'OZ', 'O1', 'O2'
]

SECOND_BATCH_CHANNEL = [
    'FP1', 'FP2', 'FZ', 'F3', 'F4', 'F7', 'F8', 'FC1', 'FC2', 'FC5', 'FC6',
    'CZ', 'C3', 'C4', 'T7', 'T8', 'CP1', 'CP2', 'CP5', 'CP6', 'PZ', 'P3', 'P4',
    'P7', 'P8', 'PO3', 'PO4', 'OZ', 'O1', 'O2', 'A2', 'A1'
]

FIRST_TO_SECOND_ORDER = [
    FIRST_BATCH_CHANNEL.index(c) for c in SECOND_BATCH_CHANNEL
]

VALENCE_DICT = {
    1: -1,  # negative
    2: -1,
    3: -1,
    4: -1,
    5: -1,
    6: -1,
    7: -1,
    8: -1,
    9: -1,
    10: -1,
    11: -1,
    12: -1,
    13: 0,  # neutral
    14: 0,
    15: 0,
    16: 0,
    17: 1,  # positive
    18: 1,
    19: 1,
    20: 1,
    21: 1,
    22: 1,
    23: 1,
    24: 1,
    25: 1,
    26: 1,
    27: 1,
    28: 1
}

EMOTION_DICT = {
    1: 0,  # anger	
    2: 0,
    3: 0,
    4: 1,  # disgust
    5: 1,
    6: 1,
    7: 2,  # fear
    8: 2,
    9: 2,
    10: 3,  # sadness
    11: 3,
    12: 3,
    13: 4,  # neutral
    14: 4,
    15: 4,
    16: 4,
    17: 5,  # amusement
    18: 5,
    19: 5,
    20: 6,  # excitation
    21: 6,
    22: 6,
    23: 7,  # happy
    24: 7,
    25: 7,
    26: 8,  # warmth
    27: 8,
    28: 8
}


def transform_producer(file_name: str, root_path: str, chunk_size: int,
                       overlap: int, channel_num: int,
                       before_trial: Union[Callable, None],
                       transform: Union[List[Callable], Callable,
                                        None], after_trial: Union[Callable,
                                                                  None],
                       reorder: bool, write_info_fn: Callable, queue: Queue):

    subject = file_name  # subject (54)
    samples = joblib.load(
        os.path.join(root_path, file_name,
                     f'{file_name}.pkl'))  # channel(33), timestep(n*250)

    events = samples[-1]
    if reorder:
        samples = samples.take(FIRST_TO_SECOND_ORDER, axis=0)

    # initial pointers
    trial_id = 0
    write_pointer = 0

    video_id = None
    start_at = None
    end_at = None

    # loop for each trial
    for i, event in enumerate(events):

        if event in list(range(1, 29)):
            # Video events 1-28: Different events correspond to different experimental video materials
            video_id = event
        elif event == 240:
            # Current trial video start event 240: This event appears 0.1s after the video event, indicating that the video starts to play.
            start_at = i
        elif event == 241:
            # Current trial video end event 241: This event indicates that the video ends playing. Block end event 243: This event indicates the end of the block.
            end_at = i
            assert (not video_id is None) and (
                not start_at is None
            ), f'Parse event fail for trial {trial_id} with video_id={video_id}, start_at={start_at}, end_at={end_at}!'

            trial_meta_info = {
                'trial_id': trial_id,
                'video_id': video_id,
                'subject_id': subject,
                'valence': VALENCE_DICT[video_id],
                'emotion': EMOTION_DICT[video_id],
            }

            cur_start_at = start_at
            if chunk_size <= 0:
                chunk_size = end_at - start_at
            cur_end_at = cur_start_at + chunk_size
            step = chunk_size - overlap

            trial_queue = []

            if before_trial:
                samples[:channel_num, cur_start_at:end_at] = before_trial(
                    samples[:channel_num, cur_start_at:end_at])

            while cur_end_at <= end_at:
                t_eeg = samples[:channel_num, cur_start_at:cur_end_at]
                if not transform is None:
                    t_eeg = transform(eeg=t_eeg)['eeg']

                clip_id = f'{subject}_{write_pointer}'
                if after_trial:
                    trial_queue.append({'eeg': t_eeg, 'key': clip_id})
                else:
                    queue.put({'eeg': t_eeg, 'key': clip_id})
                write_pointer += 1

                record_info = {
                    'start_at': cur_start_at,
                    'end_at': cur_end_at,
                    'clip_id': clip_id
                }
                record_info.update(trial_meta_info)
                write_info_fn(record_info)

                cur_start_at = cur_start_at + step
                cur_end_at = cur_start_at + chunk_size

            if len(trial_queue) and after_trial:
                trial_queue = after_trial(trial_queue)
                for obj in trial_queue:
                    assert 'eeg' in obj and 'key' in obj, 'after_trial must return a list of dictionaries, where each dictionary corresponds to an EEG sample, containing `eeg` and `key` as keys.'
                    queue.put(obj)

            # prepare for the next trial
            trial_id += 1
            video_id = None
            start_at = None
            end_at = None


def io_consumer(write_eeg_fn, queue):
    while True:
        item = queue.get()
        if not item is None:
            eeg = item['eeg']
            key = item['key']
            write_eeg_fn(eeg, key)
        else:
            break


class SingleProcessingQueue:
    def __init__(self, write_eeg_fn):
        self.write_eeg_fn = write_eeg_fn

    def put(self, item):
        eeg = item['eeg']
        key = item['key']
        self.write_eeg_fn(eeg, key)


def bci2022_constructor(root_path: str = './2022EmotionPublic/TrainSet/',
                        chunk_size: int = 250,
                        overlap: int = 0,
                        channel_num: int = 30,
                        before_trial: Union[None, Callable] = None,
                        transform: Union[None, Callable] = None,
                        after_trial: Union[Callable, None] = None,
                        io_path: str = './io/bci2022',
                        num_worker: int = 0,
                        verbose: bool = True,
                        cache_size: int = 64 * 1024 * 1024 * 1024) -> None:
    # init IO
    meta_info_io_path = os.path.join(io_path, 'info.csv')
    eeg_signal_io_path = os.path.join(io_path, 'eeg')

    if os.path.exists(io_path) and not os.path.getsize(meta_info_io_path) == 0:
        print(
            f'The target folder already exists, if you need to regenerate the database IO, please delete the path {io_path}.'
        )
        return
    os.makedirs(io_path, exist_ok=True)

    meta_info_io_path = os.path.join(io_path, 'info.csv')
    eeg_signal_io_path = os.path.join(io_path, 'eeg')

    info_io = MetaInfoIO(meta_info_io_path)
    eeg_io = EEGSignalIO(eeg_signal_io_path, cache_size=cache_size)

    for train_set_batch_index, train_set_batch in enumerate(
        ['TrainSet_first_batch', 'TrainSet_second_batch']):
        # loop to access the first batch files
        file_list = os.listdir(os.path.join(root_path, train_set_batch))

        if verbose:
            # show process bar
            pbar = tqdm(total=len(file_list))
            pbar.set_description(f"[BCI2022 BATCH {train_set_batch_index+1}/2]")

        if num_worker > 1:
            manager = Manager()
            queue = manager.Queue(maxsize=MAX_QUEUE_SIZE)
            io_consumer_process = Process(target=io_consumer,
                                          args=(eeg_io.write_eeg, queue),
                                          daemon=True)
            io_consumer_process.start()

            partial_mp_fn = partial(
                transform_producer,
                root_path=os.path.join(root_path, train_set_batch),
                chunk_size=chunk_size,
                overlap=overlap,
                channel_num=channel_num,
                before_trial=before_trial,
                transform=transform,
                after_trial=after_trial,
                reorder=train_set_batch == 'TrainSet_first_batch',
                write_info_fn=info_io.write_info,
                queue=queue)

            for _ in Pool(num_worker).imap(partial_mp_fn, file_list):
                if verbose:
                    pbar.update(1)

            queue.put(None)

            io_consumer_process.join()
            io_consumer_process.close()

        else:
            for file_name in file_list:
                transform_producer(
                    file_name=file_name,
                    root_path=os.path.join(root_path, train_set_batch),
                    chunk_size=chunk_size,
                    overlap=overlap,
                    channel_num=channel_num,
                    before_trial=before_trial,
                    transform=transform,
                    after_trial=after_trial,
                    reorder=train_set_batch == 'TrainSet_first_batch',
                    write_info_fn=info_io.write_info,
                    queue=SingleProcessingQueue(eeg_io.write_eeg))
                if verbose:
                    pbar.update(1)

        if verbose:
            pbar.close()
            print('Please wait for the writing process to complete...')



import numpy as np

from typing import Callable, Dict, Tuple, Union

from ..functional.numpy import numpy_constructor
from .base_dataset import BaseDataset


class NumpyDataset(BaseDataset):
    r'''
    A general dataset, this class converts EEG signals and annotations in Numpy format into dataset types, and caches the generated results in a unified input and output format (IO).

    A tiny case shows the use of :obj:`NumpyDataset`:

    .. code-block:: python

        # Mock 100 EEG samples. Each EEG signal contains a signal of length 1 s at a sampling rate of 128 sampled by 32 electrodes.
        X = np.random.randn(100, 32, 128)

        # Mock 100 labels, denoting valence and arousal of subjects during EEG recording.
        y = {
            'valence': np.random.randint(10, size=100),
            'arousal': np.random.randint(10, size=100)
        }
        dataset = NumpyDataset(X=X,
                               y=y,
                               io_path=io_path,
                               offline_transform=transforms.Compose(
                                   [transforms.BandDifferentialEntropy()]),
                               online_transform=transforms.ToTensor(),
                               label_transform=transforms.Compose([
                                   transforms.Select('valence'),
                                   transforms.Binary(5.0),
                               ]),
                               num_worker=2,
                               num_samples_per_trial=50)
        print(dataset[0])
        # EEG signal (torch.Tensor[32, 4]),
        # coresponding baseline signal (torch.Tensor[32, 4]),
        # label (int)

    In particular, TorchEEG utilizes the producer-consumer model to allow multi-process data preprocessing. If your data preprocessing is time consuming, consider increasing :obj:`num_worker` for higher speedup. If running under Windows, please use the proper idiom in the main module:

    .. code-block:: python
    
        if __name__ == '__main__':
            # Mock 100 EEG samples. Each EEG signal contains a signal of length 1 s at a sampling rate of 128 sampled by 32 electrodes.
            X = np.random.randn(100, 32, 128)

            # Mock 100 labels, denoting valence and arousal of subjects during EEG recording.
            y = {
                'valence': np.random.randint(10, size=100),
                'arousal': np.random.randint(10, size=100)
            }
            dataset = NumpyDataset(X=X,
                                y=y,
                                io_path=io_path,
                                offline_transform=transforms.Compose(
                                    [transforms.BandDifferentialEntropy()]),
                                online_transform=transforms.ToTensor(),
                                label_transform=transforms.Compose([
                                    transforms.Select('valence'),
                                    transforms.Binary(5.0),
                                ]),
                                num_worker=2,
                                num_samples_per_trial=50)
            print(dataset[0])
            # EEG signal (torch.Tensor[32, 4]),
            # coresponding baseline signal (torch.Tensor[32, 4]),
            # label (int)

    Args:
        X (np.ndarray): An array in :obj:`numpy.ndarray` format representing the EEG signal samples in the dataset. The shape of the array is :obj:`[num_sample, ...]` where :obj:`num_sample` is the number of samples.
        y (dict): A dictionary that records the labels corresponding to EEG samples. The keys of the dictionary represent the names of the labels, and the values are lists of labels whose length is consistent with the EEG signal samples.
        online_transform (Callable, optional): The transformation of the EEG signals and baseline EEG signals. The input is a :obj:`np.ndarray`, and the ouput is used as the first and second value of each element in the dataset. (default: :obj:`None`)
        offline_transform (Callable, optional): The usage is the same as :obj:`online_transform`, but executed before generating IO intermediate results. (default: :obj:`None`)
        label_transform (Callable, optional): The transformation of the label. The input is an information dictionary, and the ouput is used as the third value of each element in the dataset. (default: :obj:`None`)
        before_trial (Callable, optional): The hook performed on the trial to which the sample belongs. It is performed before the offline transformation and thus typically used to implement context-dependent sample transformations, such as moving averages, etc. The input and output of this hook function should be a :obj:`np.ndarray`, whose shape is (number of EEG samples per trial, ...).
        after_trial (Callable, optional): The hook performed on the trial to which the sample belongs. It is performed after the offline transformation and thus typically used to implement context-dependent sample transformations, such as moving averages, etc. The input and output of this hook function should be a sequence of dictionaries representing a sequence of EEG samples. Each dictionary contains two key-value pairs, indexed by :obj:`eeg` (the EEG signal matrix) and :obj:`key` (the index in the database) respectively
        io_path (str): The path to generated unified data IO, cached as an intermediate result. (default: :obj:`./io/deap`)
        num_worker (str): How many subprocesses to use for data processing. (default: :obj:`0`)
        verbose (bool): Whether to display logs during processing, such as progress bars, etc. (default: :obj:`True`)
        num_samples_per_trial (str): The number of samples processed by each process. Once the specified number of samples are processed, the process will be destroyed and new processes will be created to perform new tasks. (default: :obj:`100`)
        verbose (bool): Whether to display logs during processing, such as progress bars, etc. (default: :obj:`True`)
        cache_size (int): Maximum size database may grow to; used to size the memory mapping. If database grows larger than ``map_size``, an exception will be raised and the user must close and reopen. (default: :obj:`64 * 1024 * 1024 * 1024`)
    
    '''
    def __init__(self,
                 X: np.ndarray,
                 y: Dict,
                 online_transform: Union[None, Callable] = None,
                 offline_transform: Union[None, Callable] = None,
                 label_transform: Union[None, Callable] = None,
                 before_trial: Union[None, Callable] = None,
                 after_trial: Union[Callable, None] = None,
                 io_path: str = './io/numpy',
                 num_worker: int = 0,
                 num_samples_per_trial: int = 100,
                 verbose: bool = True,
                 cache_size: int = 64 * 1024 * 1024 * 1024):
        numpy_constructor(X=X,
                          y=y,
                          before_trial=before_trial,
                          transform=offline_transform,
                          after_trial=after_trial,
                          io_path=io_path,
                          num_worker=num_worker,
                          num_samples_per_trial=num_samples_per_trial,
                          verbose=verbose,
                          cache_size=cache_size)
        super().__init__(io_path)

        self.online_transform = online_transform
        self.offline_transform = offline_transform
        self.label_transform = label_transform
        self.before_trial = before_trial
        self.after_trial = after_trial
        self.num_worker = num_worker
        self.num_samples_per_trial = num_samples_per_trial
        self.verbose = verbose
        self.cache_size = cache_size

    def __getitem__(self, index: int) -> Tuple:
        info = self.info.iloc[index].to_dict()

        eeg_index = str(info['clip_id'])
        eeg = self.eeg_io.read_eeg(eeg_index)

        signal = eeg
        label = info

        if self.online_transform:
            signal = self.online_transform(eeg=eeg)['eeg']

        if self.label_transform:
            label = self.label_transform(y=info)['y']

        return signal, label

    @property
    def repr_body(self) -> Dict:
        return dict(
            super().repr_body, **{
                'online_transform': self.online_transform,
                'offline_transform': self.offline_transform,
                'label_transform': self.label_transform,
                'before_trial': self.before_trial,
                'after_trial': self.after_trial,
                'num_worker': self.num_worker,
                'num_samples_per_trial': self.num_samples_per_trial,
                'verbose': self.verbose,
                'cache_size': self.cache_size
            })
